"""Unit tests for pypyr/config.py."""
from pathlib import Path
import sys
from unittest.mock import call, mock_open, patch

import pytest

from pypyr.config import Config
from pypyr.errors import ConfigError
from pypyr.platform import PlatformPaths

CWD = Path.cwd()
current_platform = sys.platform
is_macos: bool = current_platform == 'darwin'
is_windows: bool = current_platform == 'win32'
is_posix: bool = not (is_macos or is_windows)


@pytest.fixture
def no_envs(monkeypatch):
    """Remove PYPYR_ env variables for test duration."""
    monkeypatch.delenv('PYPYR_CMD_ENCODING', raising=False)
    monkeypatch.delenv('PYPYR_ENCODING', raising=False)
    monkeypatch.delenv('PYPYR_SKIP_INIT', raising=False)
    monkeypatch.delenv('PYPYR_CONFIG_GLOBAL', raising=False)
    monkeypatch.delenv('PYPYR_CONFIG_LOCAL', raising=False)
    monkeypatch.delenv('PYPYR_NO_CACHE', raising=False)

# region default initialization


def test_config_defaults(no_envs):
    """Default config instance with all attributes as expected."""
    config = Config()
    assert config.config_loaded_paths == []
    assert config.cwd == CWD
    assert config.is_macos is is_macos
    assert config.is_posix is is_posix
    assert config.is_windows is is_windows
    assert config.platform == current_platform
    assert config.platform_paths is None
    assert config.pyproject_toml is None
    assert config.skip_init is False

    assert config.no_cache is False

    assert config.vars == {}
    assert config.shortcuts == {}

    assert config.json_ascii is False
    assert config.json_indent == 2
    assert config.pipelines_subdir == 'pipelines'
    assert config.log_config is None
    assert config.log_date_format == '%Y-%m-%d %H:%M:%S'
    assert config.log_notify_format == '%(message)s'
    assert config.log_detail_format == (
        '%(asctime)s %(levelname)s:%(name)s:%(funcName)s: %(message)s')

    assert config.default_backoff == 'fixed'
    assert config.default_cmd_encoding is None
    assert config.default_encoding is None
    assert config.default_loader == 'pypyr.loaders.file'
    assert config.default_group == 'steps'
    assert config.default_success_group == 'on_success'
    assert config.default_failure_group == 'on_failure'


def test_config_with_encoding(monkeypatch, no_envs):
    """Set encoding via env variable."""
    monkeypatch.setenv('PYPYR_ENCODING', 'arb')
    monkeypatch.setenv('PYPYR_CMD_ENCODING', 'arb2')
    config = Config()
    assert config.default_encoding == 'arb'
    assert config.default_cmd_encoding == 'arb2'


def test_config_with_no_cache(monkeypatch, no_envs):
    """Set no cache via env variable."""
    monkeypatch.setenv('PYPYR_NO_CACHE', '1')
    config = Config()
    assert config.no_cache is True

    monkeypatch.setenv('PYPYR_NO_CACHE', 'TRUE')
    config = Config()
    assert config.no_cache is True

    monkeypatch.setenv('PYPYR_NO_CACHE', 'true')
    config = Config()
    assert config.no_cache is True

    monkeypatch.setenv('PYPYR_NO_CACHE', '0')
    config = Config()
    assert config.no_cache is False

    monkeypatch.setenv('PYPYR_NO_CACHE', 'fAlse')
    config = Config()
    assert config.no_cache is False


def test_config_platforms(monkeypatch):
    """Boolean logic on which platform is correct for win vs mac vs posix."""
    monkeypatch.setattr('pypyr.platform.sys.platform', 'win32')

    config = Config()
    assert config.platform == 'win32'
    assert config.is_macos is False
    assert config.is_posix is False
    assert config.is_windows is True

    monkeypatch.setattr('pypyr.platform.sys.platform', 'linux')

    config = Config()
    assert config.platform == 'linux'
    assert config.is_macos is False
    assert config.is_posix is True
    assert config.is_windows is False

    monkeypatch.setattr('pypyr.platform.sys.platform', 'darwin')

    config = Config()
    assert config.platform == 'darwin'
    assert config.is_macos is True
    assert config.is_posix is False
    assert config.is_windows is False

    monkeypatch.setattr('pypyr.platform.sys.platform', 'cygwin')

    config = Config()
    assert config.platform == 'cygwin'
    assert config.is_macos is False
    assert config.is_posix is True
    assert config.is_windows is False

# endregion default initialization

# region init (heavy)


def test_config_skip_init(monkeypatch, no_envs):
    """Skip init when $PYPYR_SKIP_INIT true."""
    monkeypatch.setenv('PYPYR_SKIP_INIT', '1')
    config = Config()
    config.init()

    assert config.skip_init is True


fake_platform_paths = PlatformPaths(config_user=Path('/cu'),
                                    config_common=[Path('/cc')],
                                    data_dir_user=Path('/du'),
                                    data_dir_common=[Path('/dc')])


@patch('pypyr.platform.get_platform_paths', return_value=fake_platform_paths)
def test_config_init_no_files_found(mock_get_platform, no_envs):
    """Heavy init does not raise if no files found."""
    mock_opens = [
        OSError(),
        OSError(),
        OSError(),
        OSError()
    ]

    config = Config()

    with patch('builtins.open', side_effect=mock_opens) as mock_files:
        config.init()

    assert config.skip_init is False
    assert config.pipelines_subdir == 'pipelines'
    assert config.default_group == 'steps'
    assert config.pyproject_toml is None
    assert config.config_loaded_paths == []
    assert config.platform_paths == fake_platform_paths

    mock_get_platform.assert_called_once_with('pypyr', 'config.yaml')

    assert mock_files.mock_calls == [
        call(Path('/cc'), encoding=None),
        call(Path('/cu'), encoding=None),
        call(Path('pyproject.toml'), 'rb'),
        call(Path('pypyr-config.yaml'), encoding=None)]


@patch('pypyr.platform.get_platform_paths')
def test_config_init_all_files_found(mock_get_platform, no_envs):
    """Heavy init merges all files."""
    f1 = mock_open(read_data='pipelines_subdir: arb1\ndefault_backoff: x')
    f2 = mock_open(read_data='pipelines_subdir: arb2\ndefault_loader: y')
    f3 = mock_open(read_data='pipelines_subdir: arb3\ndefault_group: z')
    f4 = mock_open(read_data=(
                   b'arbkey = 123\n'
                   b'[tool.pypyr]\n'
                   b'pipelines_subdir = "arb4"\n'
                   b'default_success_group = "dsg"'))
    f5 = mock_open(
        read_data='pipelines_subdir: arb5\ndefault_failure_group: dsf')

    mock_opens = [
        f1.return_value,
        f2.return_value,
        f3.return_value,
        f4.return_value,
        f5.return_value
    ]

    fake_pp_multiple_common = PlatformPaths(
        config_user=Path('/cu'),
        config_common=[Path('/cc1'), Path('/cc2')],
        data_dir_user=Path('/du'),
        data_dir_common=[Path('/dc')])

    mock_get_platform.return_value = fake_pp_multiple_common

    config = Config()

    with patch('builtins.open', side_effect=mock_opens) as mock_files:
        config.init()

    assert config.skip_init is False
    assert config.pipelines_subdir == 'arb5'
    assert config.default_backoff == 'x'
    assert config.default_loader == 'y'
    assert config.default_group == 'z'
    assert config.default_success_group == 'dsg'
    assert config.default_failure_group == 'dsf'

    assert config.pyproject_toml == {
        'arbkey': 123,
        'tool': {'pypyr': {
            'pipelines_subdir': 'arb4',
            'default_success_group': 'dsg'
        }}}

    assert config.config_loaded_paths == [Path('/cc2'),
                                          Path('/cc1'),
                                          Path('/cu'),
                                          Path('pyproject.toml'),
                                          Path('pypyr-config.yaml')]

    assert config.platform_paths == fake_pp_multiple_common

    mock_get_platform.assert_called_once_with('pypyr', 'config.yaml')

    assert mock_files.mock_calls == [
        call(Path('/cc2'), encoding=None),
        call(Path('/cc1'), encoding=None),
        call(Path('/cu'), encoding=None),
        call(Path('pyproject.toml'), 'rb'),
        call(Path('pypyr-config.yaml'), encoding=None)]


@patch('pypyr.platform.get_platform_paths')
def test_config_init_all_files_found_scalars_and_dicts(mock_get_platform,
                                                       no_envs):
    """Heavy init merges all files with scalar and dict properties."""
    f1 = mock_open(read_data='pipelines_subdir: arb1\nvars:\n  a: b\n  f1: 1')
    f2 = mock_open(read_data='pipelines_subdir: arb2\nvars:\n  a: c\n  f2: 2')
    f3 = mock_open(read_data='pipelines_subdir: arb3\nvars:\n  a: d\n  f3: 3')
    f4 = mock_open(read_data=(
                   b'arbkey = 123\n'
                   b'[tool.pypyr]\n'
                   b'pipelines_subdir = "arb4"\n'
                   b'default_success_group = "dsg"\n'
                   b'[tool.pypyr.vars]\n'
                   b'a = "e"\n'
                   b'f4 = 4'))
    f5 = mock_open(
        read_data='pipelines_subdir: arb5\nvars:\n  a: f\n  f5: 5')

    mock_opens = [
        f1.return_value,
        f2.return_value,
        f3.return_value,
        f4.return_value,
        f5.return_value
    ]

    fake_pp_multiple_common = PlatformPaths(
        config_user=Path('/cu'),
        config_common=[Path('/cc1'), Path('/cc2')],
        data_dir_user=Path('/du'),
        data_dir_common=[Path('/dc')])

    mock_get_platform.return_value = fake_pp_multiple_common

    config = Config()

    with patch('builtins.open', side_effect=mock_opens) as mock_files:
        config.init()

    assert config.skip_init is False
    assert config.pipelines_subdir == 'arb5'
    assert config.default_backoff == 'fixed'
    assert config.default_loader == 'pypyr.loaders.file'
    assert config.default_group == 'steps'
    assert config.default_success_group == 'dsg'
    assert config.default_failure_group == 'on_failure'

    assert config.vars == {'a': 'f',
                           'f1': 1,
                           'f2': 2,
                           'f3': 3,
                           'f4': 4,
                           'f5': 5}

    assert config.shortcuts == {}

    assert config.pyproject_toml == {
        'arbkey': 123,
        'tool': {'pypyr': {
            'pipelines_subdir': 'arb4',
            'default_success_group': 'dsg',
            'vars': {
                'a': 'e',
                'f4': 4
            }
        }}}

    assert config.config_loaded_paths == [Path('/cc2'),
                                          Path('/cc1'),
                                          Path('/cu'),
                                          Path('pyproject.toml'),
                                          Path('pypyr-config.yaml')]

    assert config.platform_paths == fake_pp_multiple_common

    mock_get_platform.assert_called_once_with('pypyr', 'config.yaml')

    assert mock_files.mock_calls == [
        call(Path('/cc2'), encoding=None),
        call(Path('/cc1'), encoding=None),
        call(Path('/cu'), encoding=None),
        call(Path('pyproject.toml'), 'rb'),
        call(Path('pypyr-config.yaml'), encoding=None)]


@patch('pypyr.platform.get_platform_paths')
def test_config_init_all_files_only_dicts(mock_get_platform,
                                          no_envs):
    """Heavy init merges all files with only dict properties."""
    f1 = mock_open(read_data='vars:\n  a: b\n  f1: 1')
    f2 = mock_open(read_data='vars:\n  a: c\n  f2: 2\nshortcuts:\n  s1: one')
    f3 = mock_open(read_data='vars:\n  a: d\n  f3: 3\nshortcuts:\n  s2: two')
    f4 = mock_open(read_data=(
                   b'arbkey = 123\n'
                   b'[tool.pypyr.vars]\n'
                   b'a = "e"\n'
                   b'f4 = 4\n'
                   b'[tool.pypyr.shortcuts]\n'
                   b's1 = "oneupdated"\n'))
    f5 = mock_open(
        read_data='vars:\n  a: f\n  f5: 5\nshortcuts:\n  s3: three')

    mock_opens = [
        f1.return_value,
        f2.return_value,
        f3.return_value,
        f4.return_value,
        f5.return_value
    ]

    fake_pp_multiple_common = PlatformPaths(
        config_user=Path('/cu'),
        config_common=[Path('/cc1'), Path('/cc2')],
        data_dir_user=Path('/du'),
        data_dir_common=[Path('/dc')])

    mock_get_platform.return_value = fake_pp_multiple_common

    config = Config()

    with patch('builtins.open', side_effect=mock_opens) as mock_files:
        config.init()

    assert config.skip_init is False
    assert config.pipelines_subdir == 'pipelines'
    assert config.default_backoff == 'fixed'
    assert config.default_loader == 'pypyr.loaders.file'
    assert config.default_group == 'steps'
    assert config.default_failure_group == 'on_failure'

    assert config.vars == {'a': 'f',
                           'f1': 1,
                           'f2': 2,
                           'f3': 3,
                           'f4': 4,
                           'f5': 5}

    assert config.shortcuts == {'s1': 'oneupdated',
                                's2': 'two',
                                's3': 'three'}

    assert config.pyproject_toml == {
        'arbkey': 123,
        'tool': {'pypyr': {
            'vars': {
                'a': 'e',
                'f4': 4
            },
            'shortcuts': {
                's1': 'oneupdated'
            }
        }}}

    assert config.config_loaded_paths == [Path('/cc2'),
                                          Path('/cc1'),
                                          Path('/cu'),
                                          Path('pyproject.toml'),
                                          Path('pypyr-config.yaml')]

    assert config.platform_paths == fake_pp_multiple_common

    mock_get_platform.assert_called_once_with('pypyr', 'config.yaml')

    assert mock_files.mock_calls == [
        call(Path('/cc2'), encoding=None),
        call(Path('/cc1'), encoding=None),
        call(Path('/cu'), encoding=None),
        call(Path('pyproject.toml'), 'rb'),
        call(Path('pypyr-config.yaml'), encoding=None)]


@patch('pypyr.platform.get_platform_paths', return_value=fake_platform_paths)
def test_config_init_property_invalid(mock_get_platform, no_envs):
    """Heavy init raises on invalid property."""
    f1 = mock_open(read_data='arb: varb1\narb2: varb2\ndefault_backoff: x')

    mock_opens = [
        f1.return_value
    ]

    config = Config()

    with patch('builtins.open', side_effect=mock_opens) as mock_files:
        with pytest.raises(ConfigError) as err:
            config.init()

    assert mock_files.mock_calls == [call(Path('/cc'), encoding=None)]

    assert str(err.value) in ["Unexpected config props: {'arb2', 'arb'}",
                              "Unexpected config props: {'arb', 'arb2'}"]


@patch('pypyr.platform.get_platform_paths', return_value=fake_platform_paths)
def test_config_init_filename_set_by_env(mock_get_platform, monkeypatch):
    """Override config file name with $PYPYR_CONFIG_FILENAME."""
    monkeypatch.delenv('PYPYR_ENCODING', raising=False)
    monkeypatch.delenv('PYPYR_SKIP_INIT', raising=False)
    monkeypatch.delenv('PYPYR_CONFIG_GLOBAL', raising=False)
    monkeypatch.setenv('PYPYR_CONFIG_LOCAL', 'arb')
    mock_opens = [
        OSError,
        OSError(),
        OSError(),
        OSError
    ]

    config = Config()

    with patch('builtins.open', side_effect=mock_opens) as mock_files:
        config.init()

    assert config.skip_init is False
    assert config.pipelines_subdir == 'pipelines'
    assert config.default_group == 'steps'
    assert config.pyproject_toml is None
    assert config.config_loaded_paths == []
    assert config.platform_paths == fake_platform_paths

    mock_get_platform.assert_called_once_with('pypyr', 'config.yaml')

    assert mock_files.mock_calls == [
        call(Path('/cc'), encoding=None),
        call(Path('/cu'), encoding=None),
        call(Path('pyproject.toml'), 'rb'),
        call(Path('arb'), encoding=None)]


def test_config_file_env_override(monkeypatch, no_envs):
    """Heavy init with $PYPYR_CONFIG_FILE override."""
    monkeypatch.setenv('PYPYR_CONFIG_GLOBAL', '/arb/path')

    f1 = mock_open(read_data='pipelines_subdir: arb')

    mock_opens = [
        f1.return_value,
        OSError(),
        OSError()
    ]

    config = Config()

    with patch('builtins.open', side_effect=mock_opens) as mock_files:
        config.init()

    arb_path = Path('/arb/path')

    assert config.skip_init is False
    assert config.pipelines_subdir == 'arb'
    assert config.pyproject_toml is None
    assert config.config_loaded_paths == [arb_path]
    assert config.platform_paths == PlatformPaths(config_user=arb_path,
                                                  config_common=[arb_path],
                                                  data_dir_user=arb_path,
                                                  data_dir_common=[arb_path])

    assert mock_files.mock_calls == [
        call(Path('/arb/path'), encoding=None),
        call(Path('pyproject.toml'), 'rb'),
        call(Path('pypyr-config.yaml'), encoding=None)]


def test_config_file_env_override_merges_locals(monkeypatch, no_envs):
    """Heavy init with $PYPYR_CONFIG_FILE still merges ./ config files."""
    monkeypatch.setenv('PYPYR_CONFIG_GLOBAL', '/arb/path')

    f1 = mock_open(read_data='pipelines_subdir: arb')
    f2 = mock_open(read_data=(
        b'arbkey = 123\n'
                   b'[tool.pypyr]\n'
                   b'pipelines_subdir = "arb2"\n'
                   b'default_group = "stepsX"'))
    f3 = mock_open(read_data='pipelines_subdir: arb3\ndefault_backoff: xxx')

    mock_opens = [
        f1.return_value,
        f2.return_value,
        f3.return_value
    ]

    config = Config()

    with patch('builtins.open', side_effect=mock_opens) as mock_files:
        config.init()

    assert config.skip_init is False
    assert config.pipelines_subdir == 'arb3'
    assert config.default_group == 'stepsX'
    assert config.default_backoff == 'xxx'
    assert config.default_loader == 'pypyr.loaders.file'
    assert config.default_success_group == 'on_success'

    arb_path = Path('/arb/path')
    assert config.config_loaded_paths == [arb_path,
                                          Path('pyproject.toml'),
                                          Path('pypyr-config.yaml')]

    assert config.platform_paths == PlatformPaths(config_user=arb_path,
                                                  config_common=[arb_path],
                                                  data_dir_user=arb_path,
                                                  data_dir_common=[arb_path])
    assert config.pyproject_toml == {
        'arbkey': 123,
        'tool': {'pypyr': {
            'pipelines_subdir': 'arb2',
            'default_group': 'stepsX'
        }}}

    assert mock_files.mock_calls == [
        call(arb_path, encoding=None),
        call(Path('pyproject.toml'), 'rb'),
        call(Path('pypyr-config.yaml'), encoding=None)]


def test_config_file_env_override_not_found(monkeypatch, no_envs):
    """Heavy init with $PYPYR_CONFIG_FILE override raises on not found."""
    monkeypatch.setenv('PYPYR_CONFIG_GLOBAL', 'arb/path')

    config = Config()

    with patch('builtins.open', side_effect=OSError()) as f1:
        with pytest.raises(ConfigError) as err:
            config.init()

    path = Path('arb/path')
    assert str(err.value) == f'Could not open config file at {path}.'
    f1.assert_called_with(path, encoding=None)


def test_config_file_env_override_malformed(monkeypatch, no_envs):
    """Heavy init with $PYPYR_CONFIG_FILE override raises on not mapping."""
    monkeypatch.setenv('PYPYR_CONFIG_GLOBAL', 'arb/path')

    config = Config()

    with patch('builtins.open', mock_open(read_data='blah')) as f1:
        with pytest.raises(ConfigError) as err:
            config.init()

    path = Path('arb/path')
    assert str(err.value) == (
        f'Config file {path} should be a mapping (i.e a dict or table) at '
        'the top level.')
    f1.assert_called_with(path, encoding=None)

# endregion init (heavy)

# region pyproject.toml


@patch('pypyr.platform.get_platform_paths', return_value=fake_platform_paths)
def test_config_init_pyprojtoml_no_tool(mock_get_platform, no_envs):
    """Load pyproject.toml with no tool table."""
    f1 = mock_open(read_data=(
                   b'arbkey = 123\n'
                   b'[mytable]\n'
                   b'key1 = "arb4"\n'
                   b'default_success_group = "dsg"'))
    mock_opens = [
        OSError(),
        OSError(),
        f1.return_value,
        OSError()
    ]

    config = Config()

    with patch('builtins.open', side_effect=mock_opens) as mock_files:
        config.init()

    assert config.skip_init is False
    assert config.pipelines_subdir == 'pipelines'
    assert config.default_success_group == 'on_success'

    assert config.pyproject_toml == {
        'arbkey': 123,
        'mytable': {
            'key1': 'arb4',
            'default_success_group': 'dsg'
        }}

    assert config.config_loaded_paths == []

    assert config.platform_paths == fake_platform_paths

    mock_get_platform.assert_called_once_with('pypyr', 'config.yaml')

    assert mock_files.mock_calls == [
        call(Path('/cc'), encoding=None),
        call(Path('/cu'), encoding=None),
        call(Path('pyproject.toml'), 'rb'),
        call(Path('pypyr-config.yaml'), encoding=None)]


@patch('pypyr.platform.get_platform_paths', return_value=fake_platform_paths)
def test_config_init_pyprojtoml_tool_no_pypyr(mock_get_platform, no_envs):
    """Load pyproject.toml with tool table but no pypyr.tool sub-table."""
    f1 = mock_open(read_data=(
                   b'arbkey = 123\n'
                   b'[tool.arbtool]\n'
                   b'key1 = "arb4"\n'
                   b'default_success_group = "dsg"'))
    mock_opens = [
        OSError(),
        OSError(),
        f1.return_value,
        OSError()
    ]

    config = Config()

    with patch('builtins.open', side_effect=mock_opens) as mock_files:
        config.init()

    assert config.skip_init is False
    assert config.pipelines_subdir == 'pipelines'
    assert config.default_success_group == 'on_success'

    assert config.pyproject_toml == {
        'arbkey': 123,
        'tool': {'arbtool': {
            'key1': 'arb4',
            'default_success_group': 'dsg'
        }}}

    assert config.config_loaded_paths == []

    assert config.platform_paths == fake_platform_paths

    mock_get_platform.assert_called_once_with('pypyr', 'config.yaml')

    assert mock_files.mock_calls == [
        call(Path('/cc'), encoding=None),
        call(Path('/cu'), encoding=None),
        call(Path('pyproject.toml'), 'rb'),
        call(Path('pypyr-config.yaml'), encoding=None)]


@patch('pypyr.platform.get_platform_paths', return_value=fake_platform_paths)
def test_config_init_pyprojtoml_empty(mock_get_platform, no_envs):
    """Load pyproject.toml this is totally empty."""
    f1 = mock_open(read_data=(b''))
    mock_opens = [
        OSError(),
        OSError(),
        f1.return_value,
        OSError()
    ]

    config = Config()

    with patch('builtins.open', side_effect=mock_opens) as mock_files:
        config.init()

    assert config.skip_init is False
    assert config.pipelines_subdir == 'pipelines'
    assert config.default_success_group == 'on_success'

    # tomli parses to None rather than {}
    assert config.pyproject_toml is None

    assert config.config_loaded_paths == []

    assert config.platform_paths == fake_platform_paths

    mock_get_platform.assert_called_once_with('pypyr', 'config.yaml')

    assert mock_files.mock_calls == [
        call(Path('/cc'), encoding=None),
        call(Path('/cu'), encoding=None),
        call(Path('pyproject.toml'), 'rb'),
        call(Path('pypyr-config.yaml'), encoding=None)]


def test_pyproject_toml_with_raise(no_envs):
    """Raise error option when pyproject.toml not found."""
    config = Config()
    with patch('builtins.open', side_effect=OSError) as mock_files:
        with pytest.raises(ConfigError) as err:
            config.load_pyproject_toml(Path('pyproject.toml'), True)

    assert str(err.value) == 'Could not open config file at pyproject.toml.'
    assert config.skip_init is False
    assert config.pipelines_subdir == 'pipelines'
    assert config.default_success_group == 'on_success'

    assert config.pyproject_toml is None

    assert config.config_loaded_paths == []

    mock_files.assert_called_once_with(Path('pyproject.toml'), 'rb')
# endregion pyproject.toml

# region __str__


def test_config_default_str(no_envs):
    """Config object default represents as nice friendly string."""
    config = Config()
    assert str(config) == f"""WRITEABLE PROPERTIES:

default_backoff: fixed
default_cmd_encoding:
default_encoding:
default_failure_group: on_failure
default_group: steps
default_loader: pypyr.loaders.file
default_success_group: on_success
json_ascii: false
json_indent: 2
log_config:
log_date_format: '%Y-%m-%d %H:%M:%S'
log_detail_format: '%(asctime)s %(levelname)s:%(name)s:%(funcName)s: %(message)s'
log_notify_format: '%(message)s'
no_cache: false
pipelines_subdir: pipelines
shortcuts: {{}}
vars: {{}}


COMPUTED PROPERTIES:

config_loaded_paths: []
cwd: {CWD}
is_macos: {is_macos}
is_posix: {is_posix}
is_windows: {is_windows}
platform: {current_platform}
platform_paths:
pyproject_toml:
skip_init: False
"""  # noqa: 501
    # the noqa is to ignore line length.


@patch('pypyr.platform.get_platform_paths')
def test_config_all_str(mock_get_platform, no_envs):
    """Config represents all properties as string after heavy init."""
    f1 = mock_open(read_data=(
                   b'arbkey = 123\n'
                   b'[tool.pypyr]\n'
                   b'pipelines_subdir = "arb4"\n'
                   b'default_success_group = "dsg"\n'
                   b'no_cache = true\n'
                   b'[tool.pypyr.vars]\n'
                   b'a = "e"\n'
                   b'f4 = 4'))
    f2 = mock_open(
        read_data=('pipelines_subdir: arb5\nvars: \n  a: f\n  f5: 5\n'
                   'shortcuts:\n  s1: one'))

    mock_opens = [
        OSError(),
        OSError(),
        f1.return_value,
        f2.return_value
    ]

    cu = Path('/cu')
    cc = Path('/cc')
    du = Path('/du')
    dc = Path('/dc')
    dc2 = Path('/dc2')

    fake_pp = PlatformPaths(config_user=cu,
                            config_common=[cc],
                            data_dir_user=du,
                            data_dir_common=[dc, dc2])

    mock_get_platform.return_value = fake_pp

    config = Config()

    with patch('builtins.open', side_effect=mock_opens) as mock_files:
        config.init()

    assert config.config_loaded_paths == [Path('pyproject.toml'),
                                          Path('pypyr-config.yaml')]

    assert config.platform_paths == fake_pp

    assert config.no_cache is True

    mock_get_platform.assert_called_once_with('pypyr', 'config.yaml')

    assert mock_files.mock_calls == [
        call(cc, encoding=None),
        call(cu, encoding=None),
        call(Path('pyproject.toml'), 'rb'),
        call(Path('pypyr-config.yaml'), encoding=None)]

    assert str(config) == f"""WRITEABLE PROPERTIES:

default_backoff: fixed
default_cmd_encoding:
default_encoding:
default_failure_group: on_failure
default_group: steps
default_loader: pypyr.loaders.file
default_success_group: dsg
json_ascii: false
json_indent: 2
log_config:
log_date_format: '%Y-%m-%d %H:%M:%S'
log_detail_format: '%(asctime)s %(levelname)s:%(name)s:%(funcName)s: %(message)s'
log_notify_format: '%(message)s'
no_cache: true
pipelines_subdir: arb5
shortcuts:
  s1: one
vars:
  a: f
  f4: 4
  f5: 5


COMPUTED PROPERTIES:

config_loaded_paths:
  - pyproject.toml
  - pypyr-config.yaml
cwd: {CWD}
is_macos: {is_macos}
is_posix: {is_posix}
is_windows: {is_windows}
platform: {current_platform}
platform_paths:
  config_user: {cu}
  config_common:
    - {cc}
  data_dir_user: {du}
  data_dir_common:
    - {dc}
    - {dc2}
pyproject_toml:
  arbkey: 123
  tool:
    pypyr:
      pipelines_subdir: arb4
      default_success_group: dsg
      no_cache: true
      vars:
        a: e
        f4: 4
skip_init: False
"""  # noqa: 501
    # the noqa is to ignore line length.
# endregion __str__



"""formatting.py unit tests."""
from unittest.mock import Mock
import pytest
from pypyr.formatting import RecursionSpec, RecursiveFormatter

# region recursion_spec


def test_recursion_spec_empty():
    """Empty recursion spec initializes defaults."""
    r = RecursionSpec('')
    r.has_recursed = False
    r.is_flat = False
    r.is_recursive = False
    r.is_set = False
    r.format_spec = ''


def test_recursion_spec_no_match():
    """Input without special recursion indicators."""
    r = RecursionSpec('ABCD')
    r.has_recursed = False
    r.is_flat = False
    r.is_recursive = False
    r.is_set = False
    r.format_spec = 'ABCD'


def test_recursion_spec_no_match_less_than_two():
    """Short input without special recursion indicators."""
    r = RecursionSpec('a')
    r.has_recursed = False
    r.is_flat = False
    r.is_recursive = False
    r.is_set = False
    r.format_spec = 'a'


def test_recursion_spec_flat():
    """Parse flat recursion spec."""
    r = RecursionSpec('ff')
    r.has_recursed = False
    r.is_flat = True
    r.is_recursive = False
    r.is_set = True
    r.format_spec = ''


def test_recursion_spec_flat_with_extra():
    """Parse flat recursion spec with extra spec following."""
    r = RecursionSpec('ffabcd')
    r.has_recursed = False
    r.is_flat = True
    r.is_recursive = False
    r.is_set = True
    r.format_spec = 'abcd'


def test_recursion_spec_recursive():
    """Parse recursive recursion spec."""
    r = RecursionSpec('rf')
    r.has_recursed = False
    r.is_flat = False
    r.is_recursive = True
    r.is_set = True
    r.format_spec = ''


def test_recursion_spec_recursive_with_extra():
    """Parse recursive recursion spec with extra spec following."""
    r = RecursionSpec('rfabcd')
    r.has_recursed = False
    r.is_flat = False
    r.is_recursive = True
    r.is_set = True
    r.format_spec = 'abcd'

# endregion recursion_spec

# region RecursiveFormatter

# region RecursiveFormatter.format


def test_recursive_formatter_format_none():
    """Format a None."""
    formatter = RecursiveFormatter()
    assert formatter.format(None, 'a', 'b', 'c') is None


def test_recursive_formatter_format_empty():
    """Format an empty string."""
    formatter = RecursiveFormatter()
    assert formatter.format('', 'a', 'b', 'c') == ''


def test_recursive_formatter_format_no_expression():
    """Format a string sans formatting expression."""
    formatter = RecursiveFormatter()
    assert formatter.format('arb string', 'a', 'b', 'c') == 'arb string'


def test_recursive_formatter_format_args():
    """Format a string with positional args."""
    formatter = RecursiveFormatter()
    assert (formatter.format('{0} arb {1} string', 'a', 'b', 'c')
            == 'a arb b string')


def test_recursive_formatter_format_args_no_index():
    """Format a string with positional args and auto index."""
    formatter = RecursiveFormatter()
    assert (formatter.format('{} arb {} string', 'a', 'b', 'c')
            == 'a arb b string')


def test_recursive_formatter_format_kwargs():
    """Format a string with kwargs."""
    formatter = RecursiveFormatter()
    assert (formatter.format('{b} arb {c} string', a='a', b='b', c='c')
            == 'b arb c string')


def test_recursive_formatter_format_args_and_kwargs():
    """Format a string with positional args and kwargs."""
    formatter = RecursiveFormatter()
    assert (formatter.format('{} and {b} arb {c} string', 'a', b='b', c='c')
            == 'a and b arb c string')


def test_recursive_formatter_manual_to_auto():
    """Can't switch from manual to auto field numbering."""
    with pytest.raises(ValueError) as err:
        RecursiveFormatter().format('{0} {}', 'a', 'b')

    assert (str(err.value)
            == ('cannot switch from manual field specification to automatic '
                'field numbering'))


def test_recursive_formatter_auto_to_manual():
    """Can't switch from manual to auto field numbering."""
    with pytest.raises(ValueError) as err:
        RecursiveFormatter().format('{} {1}', 'a', 'b')

    assert (str(err.value)
            == ('cannot switch from manual field specification to automatic '
                'field numbering'))

# endregion RecursiveFormatter.format

# region RecursiveFormatter.vformat


def test_recursive_formatter_vformat_none():
    """None is None."""
    assert RecursiveFormatter().vformat(None, (1, 2), {'a': 'b'}) is None


def test_recursive_formatter_vformat_literal():
    """Literal only."""
    assert RecursiveFormatter().vformat('literal here',
                                        (1, 2),
                                        {'a': 'b'}
                                        ) == 'literal here'


def test_recursive_formatter_vformat_literal_end():
    """Literal at end."""
    assert RecursiveFormatter().vformat('{} literal here',
                                        (1, 2),
                                        {'a': 'b'}
                                        ) == '1 literal here'


def test_recursive_formatter_vformat_literal_start():
    """Literal at start."""
    assert RecursiveFormatter().vformat('literal here {a}',
                                        (1, 2),
                                        {'a': 'b'}
                                        ) == 'literal here b'


def test_recursive_formatter_vformat_literal_middle():
    """Literal in middle."""
    assert RecursiveFormatter().vformat('{} literal here {a}',
                                        (1, 2),
                                        {'a': 'b'}
                                        ) == '1 literal here b'


def test_recursive_formatter_vformat_no_literal():
    """No literal, only formatting expressions."""
    assert RecursiveFormatter().vformat('{0}{1}{a}',
                                        (1, 2),
                                        {'a': 'b'}
                                        ) == '12b'


def test_recursive_formatter_vformat_default():
    """Default formatting is standard flat."""
    d = {
        'one': '1',
        'two': '2 {one} 2',
        'three': '3 {two} 3'
    }
    assert RecursiveFormatter().vformat('start {three} end',
                                        None,
                                        d
                                        ) == 'start 3 {two} 3 end'


def test_recursive_formatter_vformat_default_with_formatting():
    """Default formatting is flat & standard formatting expression works."""
    d = {
        'one': '1',
        'two': '2 {one} 2',
        'three': '3 {two} 3'
    }
    assert RecursiveFormatter().vformat('start {three:+>11} end',
                                        None,
                                        d
                                        ) == 'start ++3 {two} 3 end'


def test_recursive_formatter_vformat_default_with_formatting_conversion():
    """Flat & standard formatting expression works with conversion."""
    d = {
        'one': '1',
        'two': '2 {one} 2',
        'three': [0, 1, '{two}']
    }
    assert RecursiveFormatter().vformat('start {three!s:+>17} end',
                                        None,
                                        d
                                        ) == "start ++[0, 1, '{two}'] end"


def test_recursive_formatter_vformat_flat():
    """Explicit flat formatting."""
    d = {
        'one': '1',
        'two': '2 {one} 2',
        'three': '3 {two} 3'
    }
    assert RecursiveFormatter().vformat('start {three:ff} end',
                                        None,
                                        d
                                        ) == 'start 3 {two} 3 end'


def test_recursive_formatter_vformat_recursive():
    """Explicit recursive formatting."""
    d = {
        'one': '1',
        'two': '2 {one} 2',
        'three': '3 {two} 3'
    }
    assert RecursiveFormatter().vformat('start {three:rf} end',
                                        None,
                                        d
                                        ) == 'start 3 2 1 2 3 end'


def test_recursive_formatter_vformat_rf_to_ff():
    """Explicit recursive formatting until ff, where it stops."""
    d = {
        'one': '1',
        'two': '2 {one} 2',
        'three': '3 {two:ff} 3'
    }
    assert RecursiveFormatter().vformat('start {three:rf} end',
                                        None,
                                        d
                                        ) == 'start 3 2 {one} 2 3 end'


def test_recursive_formatter_vformat_single_default():
    """Default formatting is for single expression is recursive."""
    d = {
        'one': '1',
        'two': '{one}',
        'three': '{two}'
    }
    assert RecursiveFormatter().vformat('{three}', None, d) == '1'


def test_recursive_formatter_vformat_single_flat():
    """Explicit flat formatting on single expression."""
    d = {
        'one': '1',
        'two': '{one}',
        'three': '{two}'
    }
    assert RecursiveFormatter().vformat('{three:ff}', None, d) == '{two}'


def test_recursive_formatter_vformat_single_recursive():
    """Explicit recursive formatting on single formatting."""
    d = {
        'one': '1',
        'two': '{one}',
        'three': '{two}'
    }
    assert RecursiveFormatter().vformat('{three:rf}', None, d) == '1'


def test_recursive_formatter_vformat_single_rf_to_ff():
    """Explicit recursive formatting on single formatting stops at ff."""
    d = {
        'one': '1',
        'two': '{one}',
        'three': '{two:ff}'
    }
    assert RecursiveFormatter().vformat('{three:rf}', None, d) == '{one}'


def test_recursive_formatter_vformat_single_default_to_ff():
    """Default recursive formatting on single formatting stops at ff."""
    d = {
        'one': '1',
        'two': '{one}',
        'three': '{two:ff}'
    }
    assert RecursiveFormatter().vformat('{three}', None, d) == '{one}'


def test_recursive_formatter_vformat_single_default_keep_type():
    """Default formatting for single expression is recursive & keeps type."""
    d = {
        'one': 1,
        'two': '{one}',
        'three': '{two}'
    }
    assert RecursiveFormatter().vformat('{three}', None, d) == 1


def test_recursive_formatter_vformat_single_default_with_conversion():
    """Single expression is recursive & conversion works."""
    d = {
        'one': 1,
        'two': '{one}',
        'three': '{two}'
    }
    assert RecursiveFormatter().vformat('{three!r:rf}', None, d) == '1'


def test_recursive_formatter_vformat_single_default_with_formatting():
    """Single expression is recursive & extra standard formatting works."""
    d = {
        'one': 1,
        'two': '{one}',
        'three': '{two}'
    }
    assert RecursiveFormatter().vformat('{three:rf+>3}', None, d) == '++1'


def test_recursive_formatter_vformat_with_passthrough():
    """Passthrough types do not format."""
    d = {
        'one': 1,
        'two': '{one}',
        'three': '{two}'
    }

    assert RecursiveFormatter(passthrough_types=str).vformat('{three}',
                                                             None,
                                                             d
                                                             ) == '{three}'


def test_recursive_formatter_vformat_with_passthrough_tuple():
    """Multiple passthrough types do not format."""
    d = {
        'one': 1,
        'two': '{one}',
        'three': {'{one}': '{two}'},
        'four': ['{one}', 'arb']
    }

    formatter = RecursiveFormatter(passthrough_types=(dict, list))
    assert formatter.vformat('{three}', None, d) == {'{one}': '{two}'}
    assert formatter.vformat(d['four'], None, d) == ['{one}', 'arb']


def test_recursive_formatter_vformat_with_special_type():
    """Special types call get_value."""
    special = Mock()
    special.get_value = Mock(return_value=123)

    d = {
        'one': 1,
        'two': special,
        'three': '{two}'
    }

    assert RecursiveFormatter(special_types=Mock).vformat('{three}',
                                                          None,
                                                          d
                                                          ) == 123
    special.get_value.assert_called_once_with(d)


class MyString(str):
    """Arbitrary test class."""

    def __new__(cls, p_string):
        """Create new class so works like string."""
        return str.__new__(cls, p_string)

    def get_value(self, kwargs):
        """Arbitrary kwargs."""
        return f'123 {kwargs["arbkey"]}'


class MyStringDerived(str):
    """Arbitrary class derived from MyString."""

    def get_value(self, kwargs):
        """Arbitrary kwargs."""
        return f'XXX {kwargs["arbkey"]}'


def test_recursive_formatter_vformat_with_special_tuple():
    """Multiple special types call get_value and supersedes string."""
    special = Mock()
    special.get_value = Mock(return_value=123)

    my_string = MyString('blah')

    d = {
        'one': 1,
        'two': special,
        'three': '{two}',
        'arbkey': 456,
        'my_string': my_string
    }

    format_me = {
        'k1': '{three}',
        'k2': '{my_string}'}

    formatter = RecursiveFormatter(special_types=(Mock, MyString))

    assert formatter.vformat(format_me,
                             None,
                             d
                             ) == {'k1': 123,
                                   'k2': '123 456'}
    special.get_value.assert_called_once_with(d)
    assert my_string == 'blah'


def test_recursive_formatter_iterate_list():
    """Recurse through a list."""
    special = Mock()
    special.get_value = Mock(return_value=123)

    my_string = MyString('blah')
    my_string_derived = MyStringDerived('blah derived')

    d = {
        'one': 1,
        'two': special,
        'three': '{two}',
        'arbkey': 456,
        'my_string': my_string,
        'my_string_derived': my_string_derived
    }

    repeating_item = Mock()
    repeating_item.get_value = Mock(return_value=999)

    passthrough_derived_obj = ValueError('arb')

    input_obj = [
        repeating_item,
        'literal string',
        'string {one} expression',
        repeating_item,
        passthrough_derived_obj,
        special,
        '{my_string}',
        MyStringDerived,
        set([1, 2, 3, 4, '{arbkey}']),
        [5, 6, 7],
        {'a': 'b', '{one}': MyStringDerived},
        b'\x00\x10',
        890,
        1.13
    ]

    formatter = RecursiveFormatter(passthrough_types=(Exception,
                                                      MyStringDerived),
                                   special_types=(Mock, MyString))
    out = formatter.vformat(input_obj, None, d)

    assert out == [
        999,
        'literal string',
        'string 1 expression',
        999,
        passthrough_derived_obj,
        123,
        '123 456',
        MyStringDerived,
        {1, 2, 3, 4, 456},
        [5, 6, 7],
        {'a': 'b', 1: MyStringDerived},
        b'\x00\x10',
        890,
        1.13
    ]

# endregion RecursiveFormatter.vformat

# region format_spec nesting


def test_recursive_formatter_recurse_format_spec():
    """Recurse on format_spec works."""
    d = {
        'k1': 'x',
        'k2': 123
    }
    assert RecursiveFormatter().vformat("a {k2:{k1}} b", None, d) == 'a 7b b'


def test_recursive_formatter_recurse_format_spec_double():
    """Recurse on format_spec exceeding max double replace."""
    d = {
        'align': '>',
        'fill': '+',
        'k3': 123
    }
    assert RecursiveFormatter().vformat(
        "a {k3:{fill}{align}5} b", None, d) == 'a ++123 b'


def test_recursive_formatter_recurse_format_spec_max_exceeded():
    """Recurse on format_spec exceeding max raises."""
    d = {
        'dash': '-',
        'i': 5,
        'align': '>',
        'fill': '+',
        'k3': 123
    }

    with pytest.raises(ValueError) as err:
        RecursiveFormatter().vformat("a {k3:{fill}{align}{i:{dash}}} b",
                                     None, d)

    assert str(err.value) == 'Max string recursion exceeded'


def test_recursive_format_keep_type_recurse_format_spec_max_exceeded():
    """Recurse on format_spec exceeding max raises.

    This test is a bit arb. Strictly speaking the recursion_depth check here
    won't be hit by any implemented code, because format_spec recursion goes
    through the standard _vformat and not _format_keep_type.

    But derived classes might. So leave a test in anyway.
    """
    d = {
        'dash': '-',
        'i': 5,
        'align': '>',
        'fill': '+',
        'k3': 123
    }

    with pytest.raises(ValueError) as err:
        RecursiveFormatter()._format_keep_type(
            "a {k3:{fill}{align}{i:{dash}}} b",
            None, d, set(), -1)

    assert str(err.value) == 'Max string recursion exceeded'

# endregion format_spec nesting

# region RecursiveFormatter.check_used_arguments


class UnusedArgs(RecursiveFormatter):
    """Unused Args test class."""

    def __init__(self, expected):
        """In expected is what used_args should be."""
        super().__init__()
        self.expected = expected

    def check_unused_args(self, used_args, args, kwargs):
        """Intercept used_args for test against expected."""
        assert used_args == self.expected


def test_recursive_formatter_unused_args():
    """Unused args as expected."""
    formatter = UnusedArgs({0, 1})
    assert (formatter.format('{} arb {} string', 'a', 'b', 'c')
            == 'a arb b string')


def test_recursive_formatter_unused_args_none():
    """Unused args as expected when all unused."""
    formatter = UnusedArgs(set())
    assert formatter.format('arb string', 'a', 'b', 'c') == 'arb string'


def test_recursive_formatted_unused_rf():
    """Unused keeps track of used args during recursion."""
    formatter = UnusedArgs({'k2', 'k3'})
    d = {
        'k1': 'v1',
        'k2': 'v2',
        'k3': [
            1,
            2,
            {'k3.1': '3.1 value {k2}'}
        ]
    }
    assert (formatter.vformat('a {k3[2][k3.1]:rf} b', None, d)
            == "a 3.1 value v2 b")


def test_recursive_formatted_unused_one_expression():
    """Unused keeps track of used args during recursion on 1 expression."""
    formatter = UnusedArgs({'k1', 'k2', 'k3'})
    d = {
        'k1': 'v1',
        'k2': '{k1}',
        'k3': [
            1,
            2,
            {'k3.1': '{k2}'}
        ]
    }
    assert (formatter.vformat('{k3[2][k3.1]}', None, d)
            == 'v1')

# endregion RecursiveFormatter.check_used_arguments

# endregion RecursiveFormatter



"""echo.py unit tests."""
import logging
import pytest
from pypyr.context import Context
from pypyr.errors import KeyNotInContextError
import pypyr.steps.echo

from tests.common.utils import patch_logger


def test_echo_pass():
    """Echo echoMe in context."""
    context = Context({
        'echoMe': 'test value here'})

    with patch_logger(
            'pypyr.steps.echo', logging.NOTIFY) as mock_logger_notify:
        pypyr.steps.echo.run_step(context)

    mock_logger_notify.assert_called_once_with('test value here')


def test_echo_substitutions_pass():
    """Format string interpolations should work."""
    context = Context({
        'key1': 'down the',
        'echoMe': 'piping {key1} valleys wild'})

    with patch_logger(
            'pypyr.steps.echo', logging.NOTIFY) as mock_logger_notify:
        pypyr.steps.echo.run_step(context)

    mock_logger_notify.assert_called_once_with('piping down the valleys wild')


def test_echo_number_pass():
    """An int should echo."""
    context = Context({
        'echoMe': 77})

    with patch_logger(
            'pypyr.steps.echo', logging.NOTIFY) as mock_logger_notify:
        pypyr.steps.echo.run_step(context)

    mock_logger_notify.assert_called_once_with(str(77))


def test_echo_bool_pass():
    """A bool should echo."""
    context = Context({
        'echoMe': False})

    with patch_logger(
            'pypyr.steps.echo', logging.NOTIFY) as mock_logger_notify:
        pypyr.steps.echo.run_step(context)

    mock_logger_notify.assert_called_once_with(str(False))


def test_echo_empty_context_fails():
    """Empty context fails."""
    with pytest.raises(AssertionError) as err_info:
        pypyr.steps.echo.run_step(None)

    assert str(err_info.value) == ("context must be set for "
                                   "echo. Did you set 'echoMe=text "
                                   "here'?")


def test_echo_missing_echo_me_raises():
    """Context must contain echoMe."""
    context = Context({
        'blah': 'blah blah'})
    with pytest.raises(KeyNotInContextError) as err_info:
        pypyr.steps.echo.run_step(context)

    assert str(err_info.value) == ("context['echoMe'] "
                                   "doesn't exist. It must exist for "
                                   "pypyr.steps.echo.")


def test_echo_echo_me_none_pass():
    """Echo can output None."""
    context = Context({'echoMe': None})
    with patch_logger(
            'pypyr.steps.echo', logging.NOTIFY) as mock_logger_notify:
        pypyr.steps.echo.run_step(context)

    mock_logger_notify.assert_called_once_with(str(None))



"""pypyr step that appends items to a mutable sequence, such as a list."""
import logging

from pypyr.utils.asserts import assert_key_exists, assert_key_is_truthy

logger = logging.getLogger(__name__)


def run_step(context):
    """Append item to a mutable sequence.

    Expects input:
        append:
            list (list or str): Add addMe to this mutable sequence.
            addMe (any): Add this item to the list.
            unpack (bool): Optional. Defaults False. If True, enumerate addMe
                and append each item individually.

    If append.list is a str, it refers to a key in context which contains a
    list, e.g context['my_list'] = [1, 2, 3]. If no such key exists, will
    create a list with that name and add addMe as the 1st item on the new list.

    This is an append, not an extend, unless append.unpack = True.

    If you want to add to a set, use pypyr.steps.add instead.

    Args:
        context (pypyr.context.Context): Mandatory. Context is a dictionary or
            dictionary-like.
            Context must contain key 'append'
    """
    logger.debug("started")

    context.assert_key_has_value(key='append', caller=__name__)

    step_input = context.get_formatted('append')

    assert_key_is_truthy(obj=step_input,
                         key='list',
                         caller=__name__,
                         parent='append')

    assert_key_exists(obj=step_input,
                      key='addMe',
                      caller=__name__,
                      parent='append')

    lst = step_input['list']
    add_me = step_input['addMe']
    is_extend = step_input.get('unpack', False)

    # str value means referring to a key in context rather than list instance
    if isinstance(lst, str):
        existing_sequence = context.get(lst, None)

        if existing_sequence:
            append_or_extend_list(existing_sequence, add_me, is_extend)
        else:
            # list(x) works only if x is iterable, [x] works when x != iterable
            context[lst] = list(add_me) if is_extend else [add_me]
    else:
        # anything that supports append: list, deque, array... if not is_extend
        append_or_extend_list(lst, add_me, is_extend)

    logger.debug("started")


def append_or_extend_list(lst, add_me, is_extend):
    """Append or extend list.

    Args:
        lst (list-like): The list to append/extend
        add_me (any): Item(s) to append/extend to lst
        is_extend (bool): If True does extend rather than append.

    Returns: None
    """
    if is_extend:
        lst.extend(add_me)
    else:
        lst.append(add_me)



"""fetchjson.py unit tests."""
from unittest.mock import mock_open, patch

import pytest

from pypyr.context import Context
from pypyr.errors import KeyInContextHasNoValueError, KeyNotInContextError
import pypyr.steps.fetchjson as filefetcher


def test_fetchjson_no_path_raises():
    """None path raises."""
    context = Context({
        'k1': 'v1'})

    with pytest.raises(KeyNotInContextError) as err_info:
        filefetcher.run_step(context)

    assert str(err_info.value) == ("context['fetchJson'] "
                                   "doesn't exist. It must exist for "
                                   "pypyr.steps.fetchjson.")


def test_fetchjson_empty_path_raises():
    """Empty path raises."""
    context = Context({
        'fetchJson': {
            'path': None}})

    with pytest.raises(KeyInContextHasNoValueError) as err_info:
        filefetcher.run_step(context)

    assert str(err_info.value) == ("context['fetchJson']['path'] must have a "
                                   "value for pypyr.steps.fetchjson.")


def test_json_pass(fs):
    """Relative path to json should succeed."""
    payload = """{
    "key1": "value1",
    "key2": "value2",
    "key3": "value3"
}
"""
    in_path = './tests/testfiles/test.json'

    fs.create_file(in_path, contents=payload)
    context = Context({
        'ok1': 'ov1',
        'fetchJson': {
            'path': in_path}})

    filefetcher.run_step(context)

    assert context, "context shouldn't be None"
    assert len(context) == 5, "context should have 5 items"
    assert context['ok1'] == 'ov1'
    assert 'fetchJsonPath' not in context
    assert context["key1"] == "value1", "key1 should be value2"
    assert context["key2"] == "value2", "key2 should be value2"
    assert context["key3"] == "value3", "key3 should be value2"


def test_json_pass_with_string(fs):
    """Relative path to json should succeed with string input."""
    payload = """{
    "key1": "value1",
    "key2": "value2",
    "key3": "value3"
}
"""
    in_path = './tests/testfiles/test.json'

    fs.create_file(in_path, contents=payload)

    context = Context({
        'ok1': 'ov1',
        'fetchJson': './tests/testfiles/test.json'})

    filefetcher.run_step(context)

    assert context, "context shouldn't be None"
    assert len(context) == 5, "context should have 5 items"
    assert context['ok1'] == 'ov1'
    assert 'fetchJsonPath' not in context
    assert context["key1"] == "value1", "key1 should be value2"
    assert context["key2"] == "value2", "key2 should be value2"
    assert context["key3"] == "value3", "key3 should be value2"


def test_json_pass_with_path_substitution(fs):
    """Relative path to json succeeds with string substitution on path."""
    payload = """{
    "key1": "value1",
    "key2": "value2",
    "key3": "value3"
}
"""
    in_path = './tests/testfiles/test.json'

    fs.create_file(in_path, contents=payload)

    context = Context({
        'ok1': 'ov1',
        'fileName': 'test',
        'fetchJson': {
            'path': './tests/testfiles/{fileName}.json'}})

    filefetcher.run_step(context)

    assert context, "context shouldn't be None"
    assert len(context) == 6, "context should have 6 items"
    assert context['ok1'] == 'ov1'
    assert 'fetchJsonPath' not in context
    assert context["key1"] == "value1", "key1 should be value2"
    assert context["key2"] == "value2", "key2 should be value2"
    assert context["key3"] == "value3", "key3 should be value2"


@patch('pypyr.config.config.default_encoding', new='utf-16')
def test_json_pass_with_encoding_from_config(fs):
    """Json loads with encoding from config."""
    payload = """{
    "key1": "value1",
    "key2": "value2",
    "key3": "value3"
}
"""
    in_path = './tests/testfiles/test.json'

    fs.create_file(in_path, contents=payload, encoding='utf-16')

    context = Context({
        'ok1': 'ov1',
        'fileName': 'test',
        'fetchJson': {
            'path': './tests/testfiles/{fileName}.json'}})

    filefetcher.run_step(context)

    assert context, "context shouldn't be None"
    assert len(context) == 6, "context should have 6 items"
    assert context['ok1'] == 'ov1'
    assert 'fetchJsonPath' not in context
    assert context["key1"] == "value1", "key1 should be value2"
    assert context["key2"] == "value2", "key2 should be value2"
    assert context["key3"] == "value3", "key3 should be value2"


@patch('pypyr.config.config.default_encoding', new='ascii')
def test_json_pass_with_encoding_from_input(fs):
    """Json loads with encoding from input."""
    payload = """{
    "key1": "value1",
    "key2": "value2",
    "key3": "value3"
}
"""
    in_path = './tests/testfiles/test.json'

    fs.create_file(in_path, contents=payload, encoding='utf-16')

    context = Context({
        'ok1': 'ov1',
        'fileName': 'test',
        'enc': 'utf-16',
        'fetchJson': {
            'path': './tests/testfiles/{fileName}.json',
            'encoding': '{enc}'}})

    filefetcher.run_step(context)

    assert context, "context shouldn't be None"
    assert len(context) == 7, "context should have 7 items"
    assert context['ok1'] == 'ov1'
    assert 'fetchJsonPath' not in context
    assert context["key1"] == "value1", "key1 should be value2"
    assert context["key2"] == "value2", "key2 should be value2"
    assert context["key3"] == "value3", "key3 should be value2"


def test_fetchjson_with_destination():
    """Json writes to destination key."""
    context = Context({
        'fetchJson': {
            'path': '/arb/arbfile',
            'key': 'outkey'}})

    with patch('pypyr.steps.fetchjson.open', mock_open(read_data='[1,2,3]')):
        filefetcher.run_step(context)

    assert context['outkey'] == [1, 2, 3]
    assert len(context) == 2


def test_fetchjson_with_destination_int():
    """Json writes to destination key that's not a string."""
    context = Context({
        'fetchJson': {
            'path': '/arb/arbfile',
            'key': 99}})

    with patch('pypyr.steps.fetchjson.open', mock_open(read_data='[1,2,3]')):
        filefetcher.run_step(context)

    assert context[99] == [1, 2, 3]
    assert len(context) == 2


def test_fetchjson_with_destination_formatting():
    """Json writes to destination key found by formatting expression."""
    context = Context({
        'keyhere': {'sub': ['outkey', 2, 3]},
        'fetchJson': {
            'path': '/arb/arbfile',
            'key': '{keyhere[sub][0]}'}})

    with patch('pypyr.steps.fetchjson.open', mock_open(
            read_data='{"1": 2,"2": 3}')):
        filefetcher.run_step(context)

    assert len(context) == 3
    assert context['outkey'] == {'1': 2, '2': 3}
    assert context['keyhere'] == {'sub': ['outkey', 2, 3]}


def test_fetchjson_list_fails():
    """Json describing a list rather than a dict should fail if no outkey."""
    context = Context({
        'ok1': 'ov1',
        'fetchJson': {
            'path': '/arb/arbfile'}})

    with patch('pypyr.steps.fetchjson.open', mock_open(read_data='[1,2,3]')):
        with pytest.raises(TypeError):
            filefetcher.run_step(context)



"""keys.py unit tests."""
import pypyr.parser.keys


def test_keys_args_parses_to_dict():
    """Args should return dictionary."""
    out = pypyr.parser.keys.get_parsed_context(['value 1',
                                                'value 2',
                                                'value3'])
    assert out['value 1'], "value 1 should be True"
    assert out['value 2'], "value 2 should be True"
    assert out['value3'], "value 3 should be True"
    assert len(out) == 3, "3 items expected"


def test_no_commas_string_parses_to_single_entry():
    """No commas input string should return dictionary with 1 item."""
    out = pypyr.parser.keys.get_parsed_context(['value 1 value 2 value3'])
    assert out['value 1 value 2 value3'], "value 1 should be True"
    assert len(out) == 1, "1 item expected"


def test_empty_string_empty_dict():
    """Empty input string should return empty dict."""
    out = pypyr.parser.keys.get_parsed_context(None)
    assert not out



"""Unit tests for pypyr.aio.subproc."""
import asyncio.subprocess as subprocess
import locale

import pytest

from pypyr.errors import ContextError
from pypyr.aio.subproc import Command, Commands

# region Command

# region ctor


def test_async_subproc_minimal():
    """Create Command with minimal inputs."""
    cmd = Command('arb')
    assert cmd.cmd == 'arb'
    assert cmd.is_shell is False
    assert cmd.cwd is None
    assert cmd.is_save is False
    assert cmd.is_text is False
    assert cmd.stdout is None
    assert cmd.stderr is None
    assert cmd.encoding == locale.getpreferredencoding(False)
    assert cmd.append is False
    assert cmd.results == []


def test_async_subproc_maximal():
    """Create Command with maximal inputs."""
    # is_save
    cmd = Command('arb',
                  is_shell=True,
                  cwd='cwd',
                  is_save=True,
                  is_text=True,
                  encoding='enc',
                  append=True)

    assert cmd.cmd == 'arb'
    assert cmd.is_shell is True
    assert cmd.cwd == 'cwd'
    assert cmd.is_save is True
    assert cmd.is_text is True
    assert cmd.stdout == subprocess.PIPE
    assert cmd.stderr == subprocess.PIPE
    assert cmd.encoding == 'enc'
    assert cmd.append is True
    assert cmd.results == []

    # not is_save
    cmd = Command('arb',
                  is_shell=True,
                  cwd='cwd',
                  is_save=False,
                  is_text=True,
                  stdout='stdout',
                  stderr='stderr',
                  encoding='enc',
                  append=True)

    assert cmd.cmd == 'arb'
    assert cmd.is_shell is True
    assert cmd.cwd == 'cwd'
    assert cmd.is_save is False
    assert cmd.is_text is False  # because save is False
    assert cmd.stdout == 'stdout'
    assert cmd.stderr == 'stderr'
    assert cmd.encoding == 'enc'
    assert cmd.append is True
    assert cmd.results == []


def test_async_subproc_command_no_stdout_on_save():
    """Raise when combining stdout/stdour with save."""
    #  stdout
    with pytest.raises(ContextError) as err:
        Command('arb', is_save=True, stdout='in')

    assert str(err.value) == (
        "You can't set `stdout` or `stderr` when `save` is True.")

    #  stderr
    with pytest.raises(ContextError) as err:
        Command('arb', is_save=True, stderr='out')

    assert str(err.value) == (
        "You can't set `stdout` or `stderr` when `save` is True.")

    #  stderr + stdout
    with pytest.raises(ContextError) as err:
        Command('arb', is_save=True, stdout='in', stderr='out')

    assert str(err.value) == (
        "You can't set `stdout` or `stderr` when `save` is True.")
# endregion ctor


def test_async_subproc_command_eq():
    """Command instance does equality check on all attributes."""
    c = Command('cmd')
    assert c == c
    assert Command('cmd') == Command('cmd')
    assert Command('cmd') != Command('other')

    assert Command([1, 2, 3]) == Command([1, 2, 3])
    assert Command([1, 2, 3]) != Command([11, 22, 33])

    assert Command('arb') != 'arb'

    assert Command('cmd',
                   is_shell=True,
                   cwd='cwd',
                   is_save=False,
                   is_text=True,
                   stdout='out',
                   stderr='err',
                   encoding='enc',
                   append=True) == Command('cmd',
                                           is_shell=True,
                                           cwd='cwd',
                                           is_save=False,
                                           is_text=True,
                                           stdout='out',
                                           stderr='err',
                                           encoding='enc',
                                           append=True)

    assert Command('cmd',
                   is_shell=False,
                   cwd='cwd',
                   is_save=False,
                   is_text=True,
                   stdout='out',
                   stderr='err',
                   encoding='enc',
                   append=False) != Command('cmd',
                                            is_shell=False,
                                            cwd='cwd',
                                            is_save=False,
                                            is_text=True,
                                            stdout='out',
                                            stderr='err',
                                            encoding='enc',
                                            append=True)
    assert Command('cmd',
                   is_shell=True,
                   cwd='cwd',
                   is_save=False,
                   is_text=True,
                   stdout='out',
                   stderr='err',
                   encoding='enc1',
                   append=True) != Command('cmd',
                                           is_shell=True,
                                           cwd='cwd',
                                           is_save=False,
                                           is_text=True,
                                           stdout='out',
                                           stderr='err',
                                           encoding='enc2',
                                           append=True)
    assert Command('cmd',
                   is_shell=True,
                   cwd='cwd',
                   is_save=True,
                   is_text=True,
                   encoding='enc',
                   append=True) == Command('cmd',
                                           is_shell=True,
                                           cwd='cwd',
                                           is_save=True,
                                           is_text=True,
                                           encoding='enc',
                                           append=True)

    assert [Command('one')] == [Command('one')]
    assert [Command('one')] != [Command('two')]

    # results
    cmd1 = Command('cmd')
    cmd1.results.append('one')
    cmd2 = Command('cmd')
    cmd2.results.append('one')
    assert cmd1 == cmd2
    cmd2.results.append('two')
    assert cmd1 != cmd2

# region parse_results


def test_async_subproc_command_parse_result_wrong_type():
    """Results should only contain Exception, SubprocessResult or list."""
    cmd = Command("arb")
    # this is pretty artificial - this err is there as a safety mechanism in
    # case of unforeseen future changes where maybe someone (me?!) forgets
    # results should only contain Exception, SubprocessResult or list.
    cmd.results.append(123)
    with pytest.raises(TypeError) as err:
        cmd.parse_results()

    # noqa is for line length
    assert str(err.value) == """\
123 is <class 'int'>.
It should be SubprocessResult | Exception | list[SubprocessResult | Exception]."""  # noqa: E501
# endregion parse_results

# region output_handles


def test_async_subproc_command_output_handles_none():
    """Output handles return None."""
    cmd = Command('arb')
    with cmd.output_handles() as (stdout, stderr):
        assert stdout is None
        assert stderr is None


def test_async_subproc_command_output_handles_pipe():
    """Output handles to pipe (capture)."""
    cmd = Command('arb', stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    with cmd.output_handles() as (stdout, stderr):
        assert stdout == subprocess.PIPE
        assert stderr == subprocess.PIPE


def test_async_subproc_command_output_handles_dev_null():
    """Output handles to /dev/null."""
    cmd = Command('arb', stdout='/dev/null', stderr='/dev/null')
    with cmd.output_handles() as (stdout, stderr):
        assert stdout == subprocess.DEVNULL
        assert stderr == subprocess.DEVNULL

    cmd = Command('arb', stderr='/dev/null')
    with cmd.output_handles() as (stdout, stderr):
        assert stdout is None
        assert stderr == subprocess.DEVNULL

    cmd = Command('arb', stdout='/dev/null')
    with cmd.output_handles() as (stdout, stderr):
        assert stdout == subprocess.DEVNULL
        assert stderr is None

    cmd = Command('arb', stdout='/dev/null', stderr='/dev/stdout')
    with cmd.output_handles() as (stdout, stderr):
        assert stdout == subprocess.DEVNULL
        assert stderr == subprocess.STDOUT
# endregion output_handles

# endregion Command

# region Commands


def test_async_subproc_commands_iter():
    """Commands is iterable and index-able."""
    cmds = Commands()
    assert len(cmds) == 0

    for c in cmds:
        raise AssertionError("can't iterate empty")

    cmds.append(Command('a'))
    cmds.append(Command('b'))

    assert len(cmds) == 2

    out = []
    for c in cmds:
        out.append(c)

    assert out == [Command('a'), Command('b')]

    assert cmds[0] == Command('a')
    assert cmds[1] == Command('b')
    assert cmds[1] != Command('c')


def test_async_subproc_commands_eq():
    """Command instance does equality check on all attributes."""
    c = Commands()
    assert c == c

    empty_cmds = Commands()
    assert empty_cmds == Commands()

    cmds = Commands()
    cmds.append(Command('a'))

    assert cmds != empty_cmds

    cmds2 = Commands()
    cmds2.append(Command('a'))

    assert cmds == cmds2
    cmds2.append(Command('b'))
    assert cmds != cmds2
    cmds.append(Command('b'))
    assert cmds == cmds2

    assert cmds != 123

# endregion Commands



"""errors.py unit tests."""
import pytest

from pypyr.errors import Error as PypyrError
from pypyr.errors import (
    ContextError,
    get_error_name,
    HandledError,
    KeyInContextHasNoValueError,
    KeyNotInContextError,
    LoopMaxExhaustedError,
    MultiError,
    PlugInError,
    PipelineDefinitionError,
    PipelineNotFoundError,
    PyModuleNotFoundError,
    Stop,
    StopStepGroup,
    StopPipeline,
    SubprocessError,
    ControlOfFlowInstruction,
    Call,
    Jump)


def test_get_error_name_builtin():
    """Builtin returns bare name on get_error_name."""
    assert get_error_name(ValueError('blah')) == 'ValueError'


def test_get_error_name_canonical():
    """Other error returns modulename.name on get_error_name."""
    assert get_error_name(ContextError('blah')) == 'pypyr.errors.ContextError'


def test_base_error_raises():
    """Pypyr root Error raises with correct message."""
    assert isinstance(PypyrError(), Exception)

    with pytest.raises(PypyrError) as err_info:
        raise PypyrError("this is error text right here")

    assert str(err_info.value) == "this is error text right here"


def test_context_error_raises():
    """A ContextError raises with correct message."""
    assert isinstance(ContextError(), PypyrError)

    with pytest.raises(ContextError) as err_info:
        raise ContextError("this is error text right here")

    assert str(err_info.value) == "this is error text right here"


def test_handled_error_raises():
    """A HandledError raises with correct message and with from."""
    assert isinstance(HandledError(), PypyrError)

    try:
        try:
            raise ContextError("this is error text right here")
        except ContextError as e:
            raise HandledError("handled") from e
    except Exception as err_info:
        assert str(err_info) == "handled"
        inner = err_info.__cause__
        assert isinstance(inner, ContextError)
        assert str(inner) == "this is error text right here"


def test_key_not_in_context_error_raises():
    """Key not in context error raises with correct message."""
    # confirm subclassed from pypyr root error
    assert isinstance(KeyNotInContextError(), PypyrError)
    assert isinstance(KeyNotInContextError(), ContextError)

    with pytest.raises(KeyNotInContextError) as err_info:
        raise KeyNotInContextError("this is error text right here")

    assert str(err_info.value) == "this is error text right here"


def test_key_in_context_has_no_value_error_raises():
    """Key not in context value error raises with correct message."""
    # confirm subclassed from pypyr root error
    assert isinstance(KeyInContextHasNoValueError(), PypyrError)
    assert isinstance(KeyNotInContextError(), ContextError)

    with pytest.raises(KeyInContextHasNoValueError) as err_info:
        raise KeyInContextHasNoValueError("this is error text right here")

    assert str(err_info.value) == "this is error text right here"


def test_loop_max_exhausted_error_raises():
    """A LoopMaxExhaustedError error raises with correct message."""
    # confirm subclassed from pypyr root error
    assert isinstance(LoopMaxExhaustedError(), PypyrError)

    with pytest.raises(LoopMaxExhaustedError) as err_info:
        raise LoopMaxExhaustedError("this is error text right here")

    assert str(err_info.value) == "this is error text right here"


def test_pipeline_definition_error_raises():
    """A PipelineDefinitionError error raises with correct message."""
    # confirm subclassed from pypyr root error
    assert isinstance(PipelineDefinitionError(), PypyrError)

    with pytest.raises(PipelineDefinitionError) as err_info:
        raise PipelineDefinitionError("this is error text right here")

    assert str(err_info.value) == "this is error text right here"


def test_pipeline_not_found_error_raises():
    """A PipelineNotFoundError error raises with correct message."""
    # confirm subclassed from pypyr root error
    assert isinstance(PipelineNotFoundError(), PypyrError)

    with pytest.raises(PipelineNotFoundError) as err_info:
        raise PipelineNotFoundError("this is error text right here")

    assert str(err_info.value) == "this is error text right here"


def test_plugin_error_raises():
    """Pypyr plugin error raises with correct message."""
    # confirm subclassed from pypyr root error
    assert isinstance(PlugInError(), PypyrError)

    with pytest.raises(PlugInError) as err_info:
        raise PlugInError("this is error text right here")

    assert str(err_info.value) == "this is error text right here"


def test_pymodule_not_found_error_raises():
    """A PyModuleNotFoundError error raises with correct message."""
    # confirm subclassed from pypyr root error
    assert isinstance(PyModuleNotFoundError(), PypyrError)

    with pytest.raises(PyModuleNotFoundError) as err_info:
        raise PyModuleNotFoundError("this is error text right here")

    assert str(err_info.value) == "this is error text right here"

# region MultiError


def test_multierror():
    """Raise a MultiError."""
    me = MultiError()
    assert isinstance(me, PypyrError)

    assert not me.has_errors

    assert len(me) == 0

    # empty repr
    assert repr(me) == 'MultiError()'

    for m in me:
        raise AssertionError('nothing to iterate')

    # append
    me.append(ValueError('one'))
    assert len(me) == 1
    assert me.has_errors
    assert str(me) == "An error occurred:\nValueError: one"

    me.append(SubprocessError(123, "cmd"))
    assert str(me) == (
        "Multiple error(s) occurred:\nValueError: one\n"
        + "SubprocessError: Command 'cmd' returned non-zero exit status 123.")
    assert len(me) == 2
    assert me.has_errors

    #  iterate
    out = [str(m) for m in me]
    assert out == ['one', "Command 'cmd' returned non-zero exit status 123."]

    # index getter
    assert repr(me[0]) == "ValueError('one')"
    assert me[1].returncode == 123

    # repr with errs but no msg
    assert repr(me) == (
        "MultiError(errors=[ValueError('one'), "
        + "SubprocessError(returncode=123, cmd='cmd')])")

    # extend
    me.extend([AssertionError('blah'), KeyError('ke')])
    assert len(me) == 4

    # with msg
    me = MultiError("msg")
    assert str(me) == 'msg'
    assert repr(me) == "MultiError(message='msg')"

    me.append(ValueError('one'))
    assert str(me) == "msg\nValueError: one"
    assert repr(me) == "MultiError(message='msg', errors=[ValueError('one')])"

    # raises
    with pytest.raises(MultiError) as err:
        raise me

    assert repr(err.value[0]) == "ValueError('one')"

# endregion MultiError

# region SubprocessError


def test_subprocess_error_raises():
    """Raise SubprocessError."""
    assert isinstance(SubprocessError(0, 'cmd'), PypyrError)

    with pytest.raises(SubprocessError) as err_caught:
        raise SubprocessError(0, 'cmd')

    err = err_caught.value
    assert err.returncode == 0
    assert err.cmd == 'cmd'
    assert err.stdout is None
    assert err.stderr is None


def test_subprocess_error_repr():
    """Convert to repr SubprocessError."""
    err_repr = repr(SubprocessError(-123, 'cmd'))
    rehydrated = eval(err_repr)

    assert rehydrated.returncode == -123
    assert rehydrated.cmd == 'cmd'
    assert rehydrated.stdout is None
    assert rehydrated.stderr is None

    err_repr = repr(SubprocessError(123, 'cmd', 'std out'))
    rehydrated = eval(err_repr)

    assert rehydrated.returncode == 123
    assert rehydrated.cmd == 'cmd'
    assert rehydrated.stdout == 'std out'
    assert rehydrated.stderr is None

    err_repr = repr(SubprocessError(123, 'cmd', 'std out', 'std err'))
    rehydrated = eval(err_repr)

    assert rehydrated.returncode == 123
    assert rehydrated.cmd == 'cmd'
    assert rehydrated.stdout == 'std out'
    assert rehydrated.stderr == 'std err'


def test_subprocess_error_str():
    """Convert to str SubprocessError."""
    #  known signal
    err_str = str(SubprocessError(-2, 'cmd'))
    assert err_str == "Command 'cmd' died with <Signals.SIGINT: 2>."

    #  unknown signal
    err_str = str(SubprocessError(-999, 'cmd'))
    assert err_str == "Command 'cmd' died with unknown signal 999."

    # return-code > 0
    err_str = str(SubprocessError(1, 'cmd'))
    assert err_str == "Command 'cmd' returned non-zero exit status 1."

    # return-code > 0 with stderr str.
    err_str = str(SubprocessError(1, 'cmd', stderr='blah'))
    assert err_str == (
        "Command 'cmd' returned non-zero exit status 1. stderr: blah")

    # return-code > 0 with stderr bytes.
    err_str = str(SubprocessError(1, 'cmd', stderr=b'blah'))
    assert err_str == (
        "Command 'cmd' returned non-zero exit status 1. stderr: b'blah'")

# endregion SubprocessError

# region Control of Flow Instructions


def test_stop_step_group_error_raises():
    """A StopErrorGroup error raises with correct message."""
    # confirm subclassed from pypyr root error
    assert isinstance(StopStepGroup(), Stop)

    with pytest.raises(StopStepGroup) as err_info:
        raise StopStepGroup("this is error text right here")

    assert str(err_info.value) == "this is error text right here"


def test_stop_pipeline_error_raises():
    """A StopPipeline error raises with correct message."""
    # confirm subclassed from pypyr root error
    assert isinstance(StopPipeline(), Stop)

    with pytest.raises(StopPipeline) as err_info:
        raise StopPipeline("this is error text right here")

    assert str(err_info.value) == "this is error text right here"


def test_jump_control_of_flow_instruction_raises():
    """A Jump instruction raises."""
    try:
        raise Jump(['one', 'two'], 'sg', 'fg', 'og')
    except Jump as err_info:
        assert isinstance(err_info, ControlOfFlowInstruction)
        assert err_info.groups == ['one', 'two']
        assert err_info.success_group == 'sg'
        assert err_info.failure_group == 'fg'
        assert err_info.original_config == 'og'


def test_call_control_of_flow_instruction_raises():
    """A Call instruction raises."""
    try:
        raise Call(['one', 'two'], 'sg', 'fg', 'og')
    except Call as err_info:
        assert isinstance(err_info, ControlOfFlowInstruction)
        assert err_info.groups == ['one', 'two']
        assert err_info.success_group == 'sg'
        assert err_info.failure_group == 'fg'
        assert err_info.original_config == 'og'

# endregion Control of Flow Instructions



"""Global cache for pipelines found by pypyr.loaders.file.

Map file path to file contents.

Attributes:
    file_cache: Global instance of the file loader cache.
                Use this attribute to access the cache from elsewhere.
"""

from pypyr.cache.cache import Cache

file_cache = Cache()



"""pypyr step yaml definition classes - domain specific language."""
from functools import reduce
import logging

from pypyr.config import config
from pypyr.utils.asserts import assert_key_has_value
from pypyr.utils.filesystem import (ObjectRewriter,
                                    StreamRewriter)


logger = logging.getLogger(__name__)


class FileInRewriterStep():
    """A pypyr step that represents a file in-out rewriter.

    This models a step that takes config like this:
        root_key:
            in: str/path-like, or list of str/paths. Mandatory.
            out: str/path-like. Optional.
            encoding: str. Optional. Encoding to use on both in & out.
            encodingIn: str. Optional. Encoding to use on reading In.
            encodingOut: str. Optional. Encoding to use on reading Out.

    If you do not set encoding, will use the system default, which is utf-8
    for everything except windows.

    Set 'encoding' to override system default for both in & out. Use encodingIn
    and encodingOut instead when you want different encodings for reading in
    and writing out.

    The run_step method does the actual work.
    """

    def __init__(self, name, root_key, context):
        """Initialize the FileInRewriterStep.

        The step config in the context dict looks like this:
            root_key:
                in: str/path-like, or list of str/paths. Mandatory.
                out: str/path-like. Optional.
                encoding: str. Optional. Encoding to use on both in & out.
                encodingIn: str. Optional. Encoding to use on reading In.
                encodingOut: str. Optional. Encoding to use on writing Out.
        Args:
            name: Unique name for step. Likely __name__ of calling step.
            root_key: str. Context key name where step's config is saved under.
            context: pypyr.context.Context. Look for config in this context
                     instance.

        """
        assert name, ("name parameter must exist for FileInRewriterStep.")
        assert root_key, ("root_key param must exist for FileInRewriterStep.")
        assert context, ("context param must exist for FileInRewriterStep.")
        # this way, logs output as the calling step, which makes more sense
        # to end-user than a mystery steps.dsl.blah logging output.
        self.logger = logging.getLogger(name)

        self.context = context
        context.assert_key_has_value(root_key, name)
        root_dict = context.get_formatted(root_key)
        self.formatted_root = root_dict

        assert_key_has_value(obj=root_dict,
                             key='in',
                             caller=name,
                             parent=root_key)

        self.path_in = root_dict['in']
        self.path_out = root_dict.get('out', None)

        encoding = root_dict.get('encoding', config.default_encoding)

        self.encoding_in = root_dict.get('encodingIn', encoding)
        self.encoding_out = root_dict.get('encodingOut', encoding)

    def run_step(self, rewriter):
        """Do the file in to out rewrite.

        Doesn't do anything more crazy than call files_in_to_out on the
        rewriter.

        Args:
            rewriter: pypyr.filesystem.FileRewriter instance.
        """
        assert rewriter, ("FileRewriter instance required to run "
                          "FileInRewriterStep.")
        rewriter.files_in_to_out(in_path=self.path_in, out_path=self.path_out)


class ObjectRewriterStep(FileInRewriterStep):
    """A pypyr step that uses an object representer to rewrite files in out."""

    def run_step(self, representer):
        """Do the object in-out rewrite.

        Args:
            representer: A pypyr.filesystem.ObjectRepresenter instance.

        """
        assert representer, ("ObjectRepresenter instance required to run "
                             "ObjectRewriterStep.")
        rewriter = ObjectRewriter(self.context.get_formatted_value,
                                  representer,
                                  encoding_in=self.encoding_in,
                                  encoding_out=self.encoding_out)
        super().run_step(rewriter)


class StreamRewriterStep(FileInRewriterStep):
    """A pypyr step that uses a streaming file in-out rewriter.

    This models a step that takes config like this:
        root_key:
            in: str/path-like, or list of str/paths. Mandatory.
            out: str/path-like. Optional.
            replacePairs: mandatory. Dictionary where items are:
                          'find_string': 'replace_string'

    """

    def run_step(self):
        """Do the file in-out rewrite."""
        rewriter = StreamRewriter(self.context.iter_formatted_strings,
                                  encoding_in=self.encoding_in,
                                  encoding_out=self.encoding_out)
        super().run_step(rewriter)


class StreamReplacePairsRewriterStep(FileInRewriterStep):
    """A stream rewriter step that uses replacePairs."""

    def __init__(self, name, root_key, context):
        """Initialize the FileInRewriterStep.

        The step config in the context dict looks like this:
            root_key:
                in: str/path-like, or list of str/paths. Mandatory.
                out: str/path-like. Optional.
                replacePairs: mandatory. Dictionary where items are:
                    'find_string': 'replace_string'
                encoding: str. Optional. Encoding to use on both in & out.
                encodingIn: str. Optional. Encoding to use on reading In.
                encodingOut: str. Optional. Encoding to use on writing Out.

        If you do not set encoding, will use the system default, which is utf-8
        for everything except windows.

        Set 'encoding' to override system default for both in & out. Use
        encodingIn and encodingOut instead when you want different encodings
        for reading in and writing out.

        Args:
            name: Unique name for step. Likely __name__ of calling step.
            root_key: str. Context key name where step's config is saved under.
            context: pypyr.context.Context. Look for config in this context
                     instance.

        Returns: None
        """
        super().__init__(name=name, root_key=root_key, context=context)
        root_dict = self.formatted_root

        assert_key_has_value(obj=root_dict,
                             key='replacePairs',
                             caller=name,
                             parent=root_key)

        self.replace_pairs = root_dict['replacePairs']

    def run_step(self):
        """Write in to out, replacing strings per the replace_pairs."""
        iter = StreamReplacePairsRewriterStep.iter_replace_strings(
            self.replace_pairs)
        rewriter = StreamRewriter(iter,
                                  encoding_in=self.encoding_in,
                                  encoding_out=self.encoding_out)
        super().run_step(rewriter)

    @staticmethod
    def iter_replace_strings(replacements):
        """Create a function that uses replacement pairs to process a string.

        The returned function takes an iterator and yields on each processed
        line.

        Args:
            replacements: Dict containing 'find_string': 'replace_string' pairs

        Returns:
            function with signature: iterator of strings = function(iterable)

        """
        def function_iter_replace_strings(iterable_strings):
            """Yield a formatted string from iterable_strings using generator.

            Args:
                iterable_strings: Iterable containing strings. E.g a file-like
                                  object.

            Returns:
                Yields formatted line.

            """
            for string in iterable_strings:
                yield reduce((lambda s, kv: s.replace(*kv)),
                             replacements.items(),
                             string)

        return function_iter_replace_strings



"""default.py unit tests."""
import logging
import pytest
from pypyr.context import Context
from pypyr.errors import KeyNotInContextError
import pypyr.steps.default
from tests.common.utils import patch_logger


def test_contextdefault_throws_on_empty_context():
    """Input context must exist."""
    with pytest.raises(KeyNotInContextError):
        pypyr.steps.default.run_step(Context())


def test_contextdefault_throws_on_contextdefault_missing():
    """Input defaults must exist in context."""
    with pytest.raises(KeyNotInContextError) as err_info:
        pypyr.steps.default.run_step(Context({'arbkey': 'arbvalue'}))

    assert str(err_info.value) == ("context['defaults'] "
                                   "doesn't exist. It must exist for "
                                   "pypyr.steps.default.")


def test_contextdefault_pass_no_substitutions():
    """Input context default success case with no substitutions."""
    context = Context({
        'key1': 'value1',
        'key2': 'value2',
        'key3': 'value3',
        'defaults': {
            'key2': 'value4',
            'key4': 'value5'
        }
    })

    pypyr.steps.default.run_step(context)

    assert context['key1'] == 'value1'
    assert context['key2'] == 'value2'
    assert context['key3'] == 'value3'
    assert context['key4'] == 'value5'


def test_contextdefault_pass_different_types_with_log():
    """Input contextdefault success case with substitutions of non strings."""
    context = Context({
        'k1': 33,
        'k2': 123.45,
        'k3': False,
        'defaults': {
            'kint': '{k1}',
            'kfloat': '{k2}',
            'kbool': '{k3}'
        }
    })

    with patch_logger('pypyr.steps.default', logging.INFO) as mock_logger_info:
        pypyr.steps.default.run_step(context)

    mock_logger_info.assert_called_once_with('set 3 context item defaults.')

    assert context['kint'] == 33
    assert context['k1'] == 33
    assert context['kfloat'] == 123.45
    assert context['k2'] == 123.45
    assert not context['kbool']
    assert isinstance(context['kbool'], bool)
    assert not context['k3']


def test_contextdefault_list():
    """Simple list."""
    context = Context({'ctx1': 'ctxvalue1',
                       'ctx2': 'ctxvalue2',
                       'ctx3': 'ctxvalue3',
                       'defaults': {
                           'ctx4': ['k1', 'k2', '{ctx3}', True, False, 44]
                       }})

    pypyr.steps.default.run_step(context)

    assert context['ctx1'] == 'ctxvalue1'
    assert context['ctx2'] == 'ctxvalue2'
    assert context['ctx3'] == 'ctxvalue3'
    assert context['ctx4'] == ['k1', 'k2', 'ctxvalue3', True, False, 44]



"""pypyr step that reads a file into context."""
import logging

from pypyr.config import config
from pypyr.utils.asserts import assert_keys_are_truthy

logger = logging.getLogger(__name__)


def run_step(context):
    """Load a file into context.

    Open file at path as text using locale.getpreferredencoding() or encoding,
    if specified.

    Open file at path as binary bytes if binary=True.

    Save file payload to key in context.

    All inputs support formatting expressions.

    For list of available encodings, see:
    https://docs.python.org/3/library/codecs.html#standard-encodings

    Args:
        context: pypyr.context.Context. Mandatory.
                 The following context key must exist
                - fileRead
                    - path. path-like. Path to file on disk.
                    - key. string. Write file payload to this context key.
                    - binary. boolean. Default False. Set to True to read file
                      content as bytes in binary mode.
                    - encoding. string. Defaults None (platform default,
                      usually 'utf-8').

    Returns:
        None. mutates context arg.

    Raises:
        FileNotFoundError: take a guess
        pypyr.errors.KeyNotInContextError: fileRead.path or fileRead.key
                                           missing in context.
        pypyr.errors.KeyInContextHasNoValueError: fileRead.path or fileRead.key
                                                  exists but is empty.

    """
    logger.debug("started")

    context.assert_key_has_value(key='fileRead', caller=__name__)

    file_read = context.get_formatted('fileRead')

    assert_keys_are_truthy(obj=file_read,
                           keys=('path', 'key'),
                           caller=__name__,
                           parent='fileRead')

    file_path = file_read['path']
    destination_key = file_read['key']
    mode = 'rb' if file_read.get('binary', False) else 'r'
    encoding = file_read.get('encoding', config.default_encoding)

    logger.debug("attempting to open file with mode '%s': %s", mode, file_path)

    with open(file_path, mode, encoding=encoding) as file:
        payload = file.read()

    context[destination_key] = payload

    logger.info("file read into pypyr context at key: %s", destination_key)
    logger.debug("done")



"""Module will not import due to invalid blahblah import."""
import blahblah.blah


print("I dont do much")


def arbmodinvalid_attribute():
    """Can't ever run on account of how blah doesn't exist."""
    blahblah.blah.blah()



"""Pytest configuration file.

This is a bit of a, ahem, alternative way of allowing testpy to find the pypyr
module, without having to go into your virtualenv or dealing with pip install
first. This is because pytest makes the dir this file is in the root & adds it
to sys.paths for you.

If you remove this file, pytest will fail UNLESS you `pip install -e .[dev]`.

Keeping this file means you can run pytest in an environment with all the
other dependencies installed but not pypyr itself.

What this file is actually for is per-directory fixture scopes:
http://doc.pytest.org/en/latest/example/simple.html#package-directory-level-fixtures-setups
"""



"""Context parser that returns a list from input arguments.

Takes input args (i.e separated by spaces on cli) and returns a list named
argList.

Where args are key=value pairs, returns a dictionary where each pair becomes a
dictionary element.

Don't have spaces in your values unless your really mean it. "k1=v1 ' k2'=v2"
will result in a context key name of ' k2' not 'k2'.

So cli input like this "ham eggs bacon drink=coffee", will yield context:
{ 'argList': ['ham', 'eggs', 'bacon'], 'drink': 'coffee'}
"""
# can remove __future__ once py 3.10 the lowest supported version
from __future__ import annotations
from collections.abc import Mapping
import logging

logger = logging.getLogger(__name__)

ARG_LIST_KEY = 'argList'


def get_parsed_context(args: list[str] | None) -> Mapping:
    """Create dict from combination of args & kwargs passed from cli.

    This supports the style of args that makefile does, i.e

    $ pypyr pipeline-name arg1 arg2 key1=value1 key2-"value 2"

    args go to context key `argList`. key=value pairs become context dictionary
    elements at root.

    Args:
      args: list of string. Passed from command-line invocation where:
            $ pypyr pipelinename this is the context_arg
            This would result in args == ['this', 'is', 'the', 'context_arg']

    Returns:
      dict. This dict will initialize the context for the pipeline run.
            The dict will have key `argList` with empty list [] if no args
            passed.
    """
    logger.debug("starting")
    if not args:
        logger.debug(
            "pipeline invoked without context arg set. For this\n"
            "argskwargs parser you're looking for something like:\n"
            "pypyr pipelinename arg1 arg2 k1=v1 k2=\"v 2\""
        )
        return {ARG_LIST_KEY: []}
    arg_list = []
    out: dict[str, str | list] = {}
    for a in args:
        # 1st = is separator, subsequent = just taken as part of the value
        key, sep, value = a.partition('=')
        if sep:
            out[key] = value
        else:
            arg_list.append(a)

    out[ARG_LIST_KEY] = arg_list

    return out



"""yaml.py unit tests."""
import io
import pytest
import ruamel.yaml as yamler
from pypyr.context import Context
import pypyr.yaml as pypyr_yaml


# region get_pipeline_yaml
def test_get_pipeline_yaml_simple():
    """Pipeline yaml loads with simple yaml."""
    file = io.StringIO('1: 2\n2: 3')
    pipeline = pypyr_yaml.get_pipeline_yaml(file)

    assert pipeline == {1: 2, 2: 3}


def test_get_pipeline_yaml_custom_types():
    """Pipeline yaml loads with custom types."""
    file = io.StringIO('1: !sic "{12}"\n2: !py abs(-23)\n3: !sic\n4: !py')
    pipeline = pypyr_yaml.get_pipeline_yaml(file)
    context = Context()

    assert len(pipeline) == 4
    assert pipeline[1].get_value() == '{12}'
    assert repr(pipeline[2]) == "PyString('abs(-23)')"
    assert pipeline[2].value == 'abs(-23)'
    assert pipeline[2].get_value(context) == 23
    # empty sic
    assert repr(pipeline[3]) == "SicString('')"
    assert pipeline[3].get_value({}) == ''
    # empty py
    assert repr(pipeline[4]) == "PyString('')"

    with pytest.raises(ValueError) as err:
        pipeline[4].get_value(context)
        assert str(err.value) == (
            '!py string expression is empty. It must be a valid python '
            'expression instead.')

# endregion get_pipeline_yaml

# region get_yaml_parser


def test_get_yaml_parser_safe():
    """Create yaml parser safe."""
    obj = pypyr_yaml.get_yaml_parser_safe()
    assert obj.typ == ['safe']
    assert obj.pure


def test_get_yaml_parser_roundtrip():
    """Create yaml parser roundtrip."""
    obj = pypyr_yaml.get_yaml_parser_roundtrip()
    assert obj.typ == ['rt']
    assert obj.pure
    assert obj.map_indent == 2
    assert obj.sequence_indent == 4
    assert obj.sequence_dash_offset == 2


def test_get_yaml_parser_roundtrip_context():
    """Create yaml parser roundtrip with Context representer."""
    obj = pypyr_yaml.get_yaml_parser_roundtrip_for_context()
    assert obj.typ == ['rt']
    assert obj.pure
    assert obj.map_indent == 2
    assert obj.sequence_indent == 4
    assert obj.sequence_dash_offset == 2
    assert obj.Representer.yaml_representers[Context] == (
        yamler.representer.RoundTripRepresenter.represent_dict)
# endregion get_yaml_parser ---------------------------------------



"""Load pipelines from string."""
import logging

from pypyr.pipedef import PipelineDefinition, PipelineFileInfo
from pypyr.yaml import get_pipeline_yaml

logger = logging.getLogger(__name__)


def get_pipeline_definition(pipeline_name, parent=None):
    """
    Receives a pipeline as string, parses it and returns a PipelineDefinition.

    Args:
        pipeline_name (str): pipeline as string.
        parent (Path-like): ignored

    Returns:
        PipelineDefinition describing the pipeline.
            The dict parsed from the pipeline is in its .pipeline property.
    """
    logger.debug("starting")

    pipeline = get_pipeline_yaml(pipeline_name)
    info = PipelineFileInfo(
        pipeline_name="", parent=parent, loader=__name__, path=None
    )
    definition = PipelineDefinition(pipeline=pipeline, info=info)

    logger.debug("found %d stages in pipeline.", len(definition.pipeline))
    logger.debug("done")

    return definition



"""Nested calls inside loops. These pipelines are in ./tests/pipelines/."""
from pypyr import pipelinerunner
from tests.common.utils import read_file_to_list

expected_file_name = '{0}_expected_output.txt'


def test_pipeline_nested():
    """Nested call control-of-flow works."""
    out = pipelinerunner.run('tests/pipelines/call/nested')

    assert out == {'out': [
        'begin',
        'A',
        'B',
        'C',
        'D',
        'end D',
        'end C',
        'end B',
        'end A',
        'end.'
    ]}

# region for


def test_pipeline_nested_for():
    """Nested call control-of-flow works with for loop."""
    out = pipelinerunner.run('tests/pipelines/call/nestedfor')

    assert out['out'] == [
        'begin',
        'sg1.1',
        'sg2.a',
        'sg2.b',
        'sg2.c',
        'sg1.2',
        'sg2.a',
        'sg2.b',
        'sg2.c',
        'sg1.3',
        'sg2.a',
        'sg2.b',
        'sg2.c',
        'end'
    ]


def test_pipeline_nested_for_deep():
    """Deep nested call control-of-flow works with for loop."""
    pipename = 'tests/pipelines/call/nestedfordeep'
    out = pipelinerunner.run('tests/pipelines/call/nestedfordeep')

    assert out['out'] == read_file_to_list(expected_file_name.format(pipename))


def test_pipeline_nested_for_groups_from_iterators():
    """Nested call control-of-flow works with groups set from iterators."""
    out = pipelinerunner.run(
        'tests/pipelines/call/nestedforgroupsfromiterator')

    assert out['out'] == [
        'begin',
        'sg1.1',
        'A: a',
        'B: b',
        'C: c',
        'end SG1',
        'sg1.2',
        'A: a',
        'B: b',
        'C: c',
        'end SG1',
        'sg1.3',
        'A: a',
        'B: b',
        'C: c',
        'end SG1',
        'end'
    ]


def test_pipeline_nested_for_formatted_groups():
    """Nested call control-of-flow works with groups set dynamically."""
    out = pipelinerunner.run('tests/pipelines/call/nestedforformatted')

    assert out['out'] == [
        'begin',
        'sg1.1',
        'sg2.a',
        'sg2.b',
        'sg2.c',
        'sg1.2',
        'gr2==end',
        'gr2==end',
        'gr2==end',
        'sg1.3',
        'gr2==end',
        'gr2==end',
        'gr2==end',
        'end'
    ]

# endregion for

# region while


def test_pipeline_nested_while():
    """Nested call control-of-flow works with while loop."""
    out = pipelinerunner.run('tests/pipelines/call/nestedwhile')

    assert out['out'] == [
        'begin',
        'sg1.1',
        'sg2.1',
        'sg2.2',
        'sg2.3',
        'sg1.2',
        'sg2.1',
        'sg2.2',
        'sg2.3',
        'sg1.3',
        'sg2.1',
        'sg2.2',
        'sg2.3',
        'end'
    ]


def test_pipeline_nested_while_swallow():
    """Nested call control-of-flow works with while loop swallowing errors."""
    out = pipelinerunner.run('tests/pipelines/call/nestedwhileswallow')

    assert out['out'] == [
        'begin',
        'sg1.1',
        'sg2.1',
        'sg2.2',
        'sg2.3',
        'sg1.2',
        'sg2.1',
        'sg2.2',
        'sg2.3',
        'sg1.3',
        'sg2.1',
        'sg2.2',
        'sg2.3',
        'end'
    ]


def test_pipeline_nested_while_for():
    """Nested call control-of-flow works with while loop AND foreach."""
    pipename = 'tests/pipelines/call/nestedwhilefor'

    out = pipelinerunner.run(pipename)
    assert out['out'] == read_file_to_list(expected_file_name.format(pipename))


def test_pipeline_nested_while_for_retry():
    """Nested call control-of-flow works with while AND foreach AND retry."""
    pipename = 'tests/pipelines/call/nestedwhileforretry'
    out = pipelinerunner.run(pipename)
    assert out['out'] == read_file_to_list(expected_file_name.format(pipename))

# endregion while

# region retries


def test_pipeline_nested_retries():
    """Nested call control-of-flow works with retry loop."""
    out = pipelinerunner.run('tests/pipelines/call/nestedretries')

    assert out['out'] == [
        'begin',
        'sg1.1',
        'sg2.1',
        'sg1.2',
        'sg2.1',
        'sg1.3',
        'sg2.1',
        'no err on sg1.3 Retry counter because of nesting is 1',
        'outer retry counter = 3',
        'end'
    ]
# endregion retries



"""cof.py unit tests."""
import logging

import pytest

from pypyr.context import Context
from pypyr.errors import (Call,
                          ContextError,
                          Jump,
                          KeyInContextHasNoValueError,
                          KeyNotInContextError)
from pypyr.steps.dsl.cof import control_of_flow_instruction as cof_func, switch

from tests.common.utils import patch_logger

# region control_of_flow_instruction

# cof_func(name, instruction_type, context, context_key)


def test_cof_name_required():
    """A CoF requires name."""
    with pytest.raises(AssertionError):
        cof_func(None, None, None, None)


def test_cof_context_required():
    """A CoF requires context."""
    with pytest.raises(AssertionError):
        cof_func('blah', None, None, None)


def test_cof_context_key_required():
    """A CoF requires context_key in context."""
    with pytest.raises(KeyNotInContextError) as err:
        cof_func('blah', None, Context({'a': 'b'}), 'key')

    assert str(err.value) == ("context['key'] doesn't exist. It must exist "
                              "for blah.")


def test_cof_context_key_groups_required():
    """A CoF requires context_key groups in context."""
    with pytest.raises(KeyNotInContextError) as err:
        cof_func('blah', None, Context({'key': {'a': 'b'}}), 'key')

    assert str(err.value) == ("key needs a child key 'groups', which should "
                              "be a list or a str with the step-group name(s) "
                              "you want to run. This is for step blah.")


def test_cof_context_key_groups_wrong_type():
    """A CoF requires context_key groups in context of str or dict."""
    with pytest.raises(ContextError) as err:
        cof_func('blah', None, Context({'key': False}), 'key')

    assert str(err.value) == ("key needs a child key 'groups', which should "
                              "be a list or a str with the step-group name(s) "
                              "you want to run. This is for step blah. "
                              "Instead, you've got False")


def test_cof_context_key_exists_but_none():
    """A CoF requires context_key in context."""
    with pytest.raises(KeyInContextHasNoValueError) as err:
        cof_func('blah', None, Context({'key': {'groups': None}}), 'key')

    assert str(err.value) == ("key.groups must have a value for step blah")


def test_cof_str_input():
    """Simple str becomes wrapped in groups list."""
    with pytest.raises(Call) as err:
        cof_func(name='blah',
                 instruction_type=Call,
                 context=Context({'key': 'b'}),
                 context_key='key')

    cof = err.value
    assert isinstance(cof, Call)
    assert cof.groups == ['b']
    assert not cof.success_group
    assert not cof.failure_group
    assert cof.original_config == ('key', 'b')


def test_cof_list_input():
    """List becomes the groups list."""
    with pytest.raises(Call) as err:
        cof_func(name='blah',
                 instruction_type=Call,
                 context=Context({'key': ['b', 'c']}),
                 context_key='key')

    cof = err.value
    assert isinstance(cof, Call)
    assert cof.groups == ['b', 'c']
    assert not cof.success_group
    assert not cof.failure_group
    assert cof.original_config == ('key', ['b', 'c'])


def test_cof_mutating_original_config():
    """Original config abides even when context mutates."""
    input_list = ['b', 'c']
    context = Context({'key': input_list})
    with pytest.raises(Call) as err:
        cof_func(name='blah',
                 instruction_type=Call,
                 context=context,
                 context_key='key')

    cof = err.value
    assert isinstance(cof, Call)
    assert cof.groups == ['b', 'c']
    assert not cof.success_group
    assert not cof.failure_group

    # mutate the list instance
    input_list[0] = 'changed'
    assert context['key'][0] == 'changed'

    # new list instance creates new object, original remains intact
    input_list = ['d', 'e']
    assert context['key'][0] == 'changed'
    assert cof.original_config == ('key', ['changed', 'c'])


def test_cof_dict_with_str_input():
    """Dict with simple str becomes wrapped in groups list."""
    with pytest.raises(Call) as err:
        cof_func(name='blah',
                 instruction_type=Call,
                 context=Context({'key': {'groups': 'b'}}),
                 context_key='key')

    cof = err.value
    assert isinstance(cof, Call)
    assert cof.groups == ['b']
    assert not cof.success_group
    assert not cof.failure_group
    assert cof.original_config == ('key', {'groups': 'b'})


def test_cof_dict_with_list_input():
    """Dict with simple list."""
    with pytest.raises(Call) as err:
        cof_func(name='blah',
                 instruction_type=Call,
                 context=Context({'key': {'groups': ['b', 'c']}}),
                 context_key='key')

    cof = err.value
    assert isinstance(cof, Call)
    assert cof.groups == ['b', 'c']
    assert not cof.success_group
    assert not cof.failure_group
    assert cof.original_config == ('key', {'groups': ['b', 'c']})


def test_cof_dict_with_all_args():
    """Dict with all values set."""
    with pytest.raises(Call) as err:
        with patch_logger('blah', logging.INFO) as mock_logger_info:
            cof_func(name='blah',
                     instruction_type=Call,
                     context=Context({'key': {'groups': ['b', 'c'],
                                              'success': 'sg',
                                              'failure': 'fg'}}),
                     context_key='key')

    cof = err.value
    assert isinstance(cof, Call)
    assert cof.groups == ['b', 'c']
    assert cof.success_group == 'sg'
    assert cof.failure_group == 'fg'
    assert cof.original_config == ('key', {'groups': ['b', 'c'],
                                           'success': 'sg',
                                           'failure': 'fg'})

    mock_logger_info.assert_called_once_with(
        "step blah about to hand over control with key: Will run groups: "
        "['b', 'c']  with success sg and failure fg")


def test_cof_dict_with_success_not_a_string():
    """Dict with success_group not a string should raise."""
    with pytest.raises(ContextError) as err:
        cof_func(name='blah',
                 instruction_type=Call,
                 context=Context({'key': {'groups': ['b', 'c'],
                                          'success': 1,
                                          'failure': 'fg'}}),
                 context_key='key')

    assert str(err.value) == "key.success must be a string for blah."


def test_cof_dict_with_failure_not_a_string():
    """Dict with failure_group not a string should raise."""
    with pytest.raises(ContextError) as err:
        cof_func(name='blah',
                 instruction_type=Call,
                 context=Context({'key': {'groups': ['b', 'c'],
                                          'success': '1',
                                          'failure': False}}),
                 context_key='key')

    assert str(err.value) == "key.failure must be a string for blah."


def test_cof_dict_with_all_args_formatting():
    """Dict with all values set from formatting expressions."""
    with pytest.raises(Jump) as err:
        cof_func(name='blah',
                 instruction_type=Jump,
                 context=Context({
                     'list': ['b', 'c'],
                     'sg': 'sgv',
                     'fg': 'fgv',
                     'key': {
                         'groups': '{list}',
                         'success': '{sg}',
                         'failure': '{fg}'
                     }}),
                 context_key='key')

    cof = err.value
    assert isinstance(cof, Jump)
    assert cof.groups == ['b', 'c']
    assert cof.success_group == 'sgv'
    assert cof.failure_group == 'fgv'
    assert cof.original_config == ('key', {'groups': '{list}',
                                           'success': '{sg}',
                                           'failure': '{fg}'
                                           })


def test_cof_dict_with_all_args_formatting_to_string():
    """Dict with all values set from formatting expressions to string."""
    with pytest.raises(Jump) as err:
        cof_func(name='blah',
                 instruction_type=Jump,
                 context=Context({
                     'list': 'abc',
                     'sg': 'sgv',
                     'fg': 'fgv',
                     'key': {
                         'groups': '{list}',
                         'success': '{sg}',
                         'failure': '{fg}'
                     }}),
                 context_key='key')

    cof = err.value
    assert isinstance(cof, Jump)
    assert cof.groups == ['abc']
    assert cof.success_group == 'sgv'
    assert cof.failure_group == 'fgv'
    assert cof.original_config == ('key', {'groups': '{list}',
                                           'success': '{sg}',
                                           'failure': '{fg}'
                                           })

# endregion control_of_flow_instruction

# region switch


def test_switch_no_switch_input():
    """Raise error when no switch input key on switch."""
    with pytest.raises(KeyNotInContextError) as err:
        switch(Context({'a': 'b'}), 'blah')

    assert str(err.value) == ("context['switch'] doesn't exist. It must exist "
                              "for blah.")


def test_switch_context_key_exists_but_none():
    """Raise error when switch key exists but is None."""
    with pytest.raises(KeyInContextHasNoValueError) as err:
        switch(Context({'switch': None}), 'blah')

    assert str(err.value) == (
        "context['switch'] must have a value for blah.")


def test_switch_missing_case():
    """Switch raises if list element found with no case key."""
    with pytest.raises(KeyNotInContextError) as err:
        switch(context=Context({
            'list': 'sg1',
            'case1': True,
            'fg': 'fgv',
            'switch': [
                {'caseX': '{case1}', 'call': '{list}'},
                {'case': '{case1}', 'call': 'sg2'},
            ]}),
            name='blah')

    assert str(err.value) == "'case' not found in `switch` index 0."


def test_switch_missing_case_because_out_of_order_default():
    """Switch raises if default not last."""
    with pytest.raises(KeyNotInContextError) as err:
        switch(context=Context({
            'list': 'sg1',
            'case1': False,
            'fg': 'fgv',
            'switch': [
                {'case': '{case1}', 'call': '{list}'},
                {'default': 'blah'},
                {'case': '{case1}', 'call': 'sg2'},
            ]}),
            name='blah')

    assert str(err.value) == "'case' not found in `switch` index 1."


def test_switch_empty_case():
    """Switch allows empty or None case."""
    switch(context=Context({
        'list': 'sg1',
        'case1': True,
        'fg': 'fgv',
        'switch': [
                {'case': '', 'call': '{list}'},
                {'case': None, 'call': 'sg2'},
        ]}),
        name='blah')


def test_switch_missing_call():
    """Switch raises if list element found with no call key."""
    with pytest.raises(KeyNotInContextError) as err:
        switch(context=Context({
            'list': 'sg1',
            'case1': False,
            'fg': 'fgv',
            'switch': [
                {'case': '{case1}', 'call': '{list}'},
                {'case': '{case1}', 'callX': 'sg2'},
            ]}),
            name='blah')

    assert str(err.value) == "'call' not found in `switch` index 1."


def test_switch_call_no_value():
    """Switch raises if list element found with call key with no value."""
    with pytest.raises(KeyInContextHasNoValueError) as err:
        switch(context=Context({
            'list': 'sg1',
            'case1': False,
            'fg': 'fgv',
            'switch': [
                {'case': '{case1}', 'call': '{list}'},
                {'case': '{case1}', 'call': ''},
            ]}),
            name='blah')

    assert str(err.value) == (
        "'call' does not have a value at `switch` index 1.")


def test_switch_first():
    """Switch only runs first if case True."""
    with pytest.raises(Call) as err:
        switch(context=Context({
            'list': 'sg1',
            'case1': True,
            'fg': 'fgv',
            'switch': [
                {'case': '{case1}', 'call': '{list}'},
                {'case': '{case1}', 'call': 'sg2'},
            ]}),
            name='blah')

    cof = err.value
    assert isinstance(cof, Call)
    assert cof.groups == ['sg1']
    assert cof.success_group is None
    assert cof.failure_group is None
    assert cof.original_config == ('switch', [
        {'case': '{case1}', 'call': '{list}'},
        {'case': '{case1}', 'call': 'sg2'},
    ])


def test_switch_last():
    """Switch only runs last if case True."""
    with pytest.raises(Call) as err:
        switch(context=Context({
            'list': 'sg1',
            'case1': False,
            'case2': True,
            'fg': 'fgv',
            'switch': [
                {'case': '{case1}', 'call': '{list}'},
                {'case': '{case2}', 'call': 'sg2'},
            ]}),
            name='blah')

    cof = err.value
    assert isinstance(cof, Call)
    assert cof.groups == ['sg2']
    assert cof.success_group is None
    assert cof.failure_group is None
    assert cof.original_config == ('switch', [
        {'case': '{case1}', 'call': '{list}'},
        {'case': '{case2}', 'call': 'sg2'},
    ])


def test_switch_default():
    """Switch only runs default if no case True."""
    with pytest.raises(Call) as err:
        switch(context=Context({
            'list': 'sg1',
            'def': 'sg3',
            'case1': False,
            'case2': False,
            'fg': 'fgv',
            'switch': [
                {'case': '{case1}', 'call': '{list}'},
                {'case': '{case2}', 'call': 'sg2'},
                {'default': '{def}'}
            ]}),
            name='blah')

    cof = err.value
    assert isinstance(cof, Call)
    assert cof.groups == ['sg3']
    assert cof.success_group is None
    assert cof.failure_group is None
    assert cof.original_config == ('switch', [
        {'case': '{case1}', 'call': '{list}'},
        {'case': '{case2}', 'call': 'sg2'},
        {'default': '{def}'}
    ])


def test_switch_default_list_input():
    """Switch only runs default if no case True with >1 groups."""
    with pytest.raises(Call) as err:
        switch(context=Context({
            'list': 'sg1',
            'def': 'sg3',
            'case1': False,
            'case2': False,
            'fg': 'fgv',
            'switch': [
                {'case': '{case1}', 'call': '{list}'},
                {'case': '{case2}', 'call': 'sg2'},
                {'default': ['{def}', 'sg4']}
            ]}),
            name='blah')

    cof = err.value
    assert isinstance(cof, Call)
    assert cof.groups == ['sg3', 'sg4']
    assert cof.success_group is None
    assert cof.failure_group is None
    assert cof.original_config == ('switch', [
        {'case': '{case1}', 'call': '{list}'},
        {'case': '{case2}', 'call': 'sg2'},
        {'default': ['{def}', 'sg4']}
    ])


def test_switch_default_all_call_inputs():
    """Switch only runs default if no case True with all call inputs."""
    with pytest.raises(Call) as err:
        switch(context=Context({
            'list': 'sg1',
            'def': 'sg3',
            'case1': False,
            'case2': False,
            'sg': 'sgv',
            'fg': 'fgv',
            'switch': [
                {'case': '{case1}', 'call': '{list}'},
                {'case': '{case2}', 'call': 'sg2'},
                {'default': {
                    'groups': ['{def}', 'sg4'],
                    'success': '{sg}',
                    'failure': '{fg}'
                }}
            ]}),
            name='blah')

    cof = err.value
    assert isinstance(cof, Call)
    assert cof.groups == ['sg3', 'sg4']
    assert cof.success_group == 'sgv'
    assert cof.failure_group == 'fgv'
    assert cof.original_config == ('switch', [
        {'case': '{case1}', 'call': '{list}'},
        {'case': '{case2}', 'call': 'sg2'},
        {'default': {
            'groups': ['{def}', 'sg4'],
            'success': '{sg}',
            'failure': '{fg}'
        }}
    ])


def test_switch_case_all_call_inputs():
    """Switch runs case in the middle with all call inputs."""
    with pytest.raises(Call) as err:
        switch(context=Context({
            'list': 'sg1',
            'def': 'sg3',
            'sg2_input': 'sg2',
            'case1': False,
            'case2': True,
            'sg': 'sgv',
            'fg': 'fgv',
            'switch': [
                {'case': '{case1}', 'call': '{list}'},
                {'case': '{case2}', 'call': {
                    'groups': '{sg2_input}',
                    'success': '{sg}',
                    'failure': '{fg}'}},
                {'default': {
                    'groups': ['{def}', 'sg4'],
                    'success': '{sg}d',
                    'failure': '{fg}d'
                }}
            ]}),
            name='blah')

    cof = err.value
    assert isinstance(cof, Call)
    assert cof.groups == ['sg2']
    assert cof.success_group == 'sgv'
    assert cof.failure_group == 'fgv'
    assert cof.original_config == ('switch', [
        {'case': '{case1}', 'call': '{list}'},
        {'case': '{case2}', 'call': {
            'groups': '{sg2_input}',
            'success': '{sg}',
            'failure': '{fg}'}},
        {'default': {
            'groups': ['{def}', 'sg4'],
            'success': '{sg}d',
            'failure': '{fg}d'
        }}
    ])


def test_switch_none_no_default():
    """Switch does nothing if no cases True and no default set."""
    switch(context=Context({
        'list': 'sg1',
        'case1': False,
        'case2': False,
        'fg': 'fgv',
        'switch': [
                {'case': '{case1}', 'call': '{list}'},
                {'case': '{case2}', 'call': 'sg2'},
        ]}),
        name='blah')

# endregion switch



"""Unit tests for pypyr.parser.argskwargs."""

from pypyr.parser.argskwargs import get_parsed_context


def test_argskwargs_empty_input():
    """Parse empty inputs."""
    assert get_parsed_context([]) == {'argList': []}
    assert get_parsed_context(None) == {'argList': []}


def test_argskwargs_args_only():
    """Parse when only args."""
    assert get_parsed_context(['a']) == {'argList': ['a']}
    assert get_parsed_context(['a', 'b b', 'c']) == {'argList':
                                                     ['a', 'b b', 'c']}


def test_argskwargs_kwargs_only():
    """Parse when only kwargs."""
    assert get_parsed_context(['k1=value 1']) == {'argList': [],
                                                  'k1': 'value 1'}
    assert get_parsed_context(['k1=value 1', 'k2=v2=3']) == {'argList': [],
                                                             'k1': 'value 1',
                                                             'k2': 'v2=3'}


def test_argskwargs_args_and_kwargs():
    """Parse when args + kwargs."""
    assert get_parsed_context(['a', 'b', 'c d',
                               'k1="value 1"', 'k2=v2']) == {
        'argList': ['a', 'b', 'c d'],
        'k1': '"value 1"',
        'k2': 'v2'}
